import collections
import typing
import uqbar.graphs
from abjad import enums
from abjad import exceptions
from abjad.top.inspect import inspect
from abjad.top.iterate import iterate
from abjad.top.select import select
from abjad.utilities.Duration import Duration
from .Component import Component
from .Selection import Selection


class Container(Component):
    r"""
    Container.

    ..  container:: example

        Intializes from string:

        >>> container = abjad.Container("c'4 e'4 d'4 e'8 f'8")
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                e'4
                d'4
                e'8
                f'8
            }

    ..  container:: example

        Intializes from components:

        >>> notes = [
        ...     abjad.Note("c'4"),
        ...     abjad.Note("e'4"),
        ...     abjad.Note("d'4"),
        ...     abjad.Note("e'8"),
        ...     abjad.Note("f'8"),
        ...     ]
        >>> container = abjad.Container(notes)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                e'4
                d'4
                e'8
                f'8
            }

    ..  container:: example

        Intializes from selections:

        >>> notes = [
        ...     abjad.Note("c'4"),
        ...     abjad.Note("e'4"),
        ...     abjad.Note("d'4"),
        ...     abjad.Note("e'8"),
        ...     abjad.Note("f'8"),
        ...     ]
        >>> selection = abjad.select(notes)
        >>> container = abjad.Container(selection)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                e'4
                d'4
                e'8
                f'8
            }

    ..  container:: example

        Intializes from mixed components and selections:

        >>> items = [
        ...     abjad.Note("c'4"),
        ...     abjad.select(abjad.Note("e'4")),
        ...     abjad.select(abjad.Note("d'4")),
        ...     abjad.Note("e'8"),
        ...     abjad.Note("f'8"),
        ...     ]
        >>> container = abjad.Container(items)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                e'4
                d'4
                e'8
                f'8
            }

    ..  container:: example

        Containers are iterables:

        >>> import collections
        >>> container = abjad.Container("c'4 e'4 d'4 e'8 f'8")
        >>> isinstance(container, collections.Iterable)
        True

    ..  container:: example

        Containers are not sequences because containers do not implement
        reverse:

        >>> container = abjad.Container("c'4 e'4 d'4 e'8 f'8")
        >>> isinstance(container, collections.Sequence)
        False

    ..  container:: example

        Formatting positions contributions strictly one-per-line:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Articulation('^'), staff[0])
        >>> abjad.attach(abjad.Markup('Allegro', direction=abjad.Up), staff[0])
        >>> abjad.attach(abjad.StemTremolo(), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            :16
            -\marcato
            ^ \markup { Allegro }
            d'4
            e'4
            f'4
        }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Containers'

    __slots__ = (
        '_identifier',
        '_components',
        '_formatter',
        '_named_children',
        '_name',
        '_is_simultaneous',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        components=None,
        identifier: str = None,
        is_simultaneous: bool = None,
        name: str = None,
        ) -> None:
        components = components or []
        Component.__init__(self)
        self._named_children: dict = {}
        self._is_simultaneous = None
        self._initialize_components(components)
        self.identifier = identifier
        self.is_simultaneous = is_simultaneous
        # NOTE: name must be set up *after* parent
        self._name = None
        self.name = name

    ### SPECIAL METHODS ###

    def __contains__(self, argument) -> bool:
        """
        Is true when ``argument`` appears in container.
        """
        if isinstance(argument, str):
            return argument in self._named_children
        else:
            for component in self.components:
                if component is argument:
                    return True
            else:
                return False

    def __delitem__(self, i):
        r"""
        Deletes components(s) at index ``i`` in container.

        ..  container:: example

            Deletes first tuplet in voice:

            >>> voice = abjad.Voice()
            >>> voice.append(abjad.Tuplet((2, 3), "c'4 d'4 e'4"))
            >>> voice.append(abjad.Tuplet((2, 3), "e'4 d'4 c'4"))
            >>> leaves = abjad.select(voice).leaves()
            >>> abjad.attach(abjad.Slur(), leaves)
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(voice)
                \new Voice
                {
                    \times 2/3 {
                        c'4
                        (
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        e'4
                        d'4
                        c'4
                        )
                    }
                }

            >>> tuplet_1 = voice[0]
            >>> del(voice[0])

            First tuplet no longer appears in voice:

                >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(voice)
                \new Voice
                {
                    \times 2/3 {
                        e'4
                        (
                        d'4
                        c'4
                        )
                    }
                }

            >>> abjad.inspect(voice).is_well_formed()
            True

            First tuplet is no longer slurred but is still well-formed:

            >>> abjad.show(tuplet_1) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet_1)
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }

            >>> abjad.inspect(tuplet_1).is_well_formed()
            True

        Withdraws component(s) from crossing spanners.

        Preserves spanners that component(s) cover(s).

        Returns none.
        """
        components = self[i]
        if not isinstance(components, Selection):
            components = select([components])
        if not self.is_simultaneous:
            components._withdraw_from_crossing_spanners()
        components._set_parents(None)

    def __getitem__(self, argument) -> typing.Union[Component, Selection]:
        """
        Gets item or slice identified by ``argument``.

        Traverses top-level items only.
        """
        if isinstance(argument, int):
            return self.components.__getitem__(argument)
        elif isinstance(argument, slice) and not self.is_simultaneous:
            return select(self.components.__getitem__(argument))
        elif isinstance(argument, slice) and self.is_simultaneous:
            return select(self.components.__getitem__(argument))
        elif isinstance(argument, str):
            if argument not in self._named_children:
                message = 'can not find component named {!r}.'
                message = message.format(argument)
                raise ValueError(message)
            elif 1 < len(self._named_children.__getitem__(argument)):
                message = 'multiple components named {!r}.'
                message = message.format(argument)
                raise ValueError(message)
            return self._named_children.__getitem__(argument)[0]
        message = 'can not get container at {!r}.'
        message = message.format(argument)
        raise ValueError(message)

    def __getnewargs__(self) -> tuple:
        """
        Gets new container arguments.
        """
        return [], self.identifier, self.is_simultaneous, self.name

    def __graph__(self, spanner=None, **keywords):
        """
        Graphviz graph representation of container.

        Returns Graphviz graph.
        """
        def recurse(component, leaf_cluster):
            component_node = component._as_graphviz_node()
            node_mapping[component] = component_node
            node_order = [component_node.name]
            if isinstance(component, Container):
                graph.append(component_node)
                this_leaf_cluster = uqbar.graphs.Graph(
                    name=component_node.name,
                    attributes={
                        'color': 'grey75',
                        'penwidth': 2,
                        },
                    )
                all_are_leaves = True
                pending_node_order = []
                for child in component:
                    if not isinstance(child, abjad.Leaf):
                        all_are_leaves = False
                    child_node, child_node_order = recurse(
                        child, this_leaf_cluster)
                    pending_node_order.extend(child_node_order)
                    edge = uqbar.graphs.Edge()
                    edge.attach(component_node, child_node)
                if all_are_leaves:
                    pending_node_order.reverse()
                node_order.extend(pending_node_order)
                if len(this_leaf_cluster):
                    leaf_cluster.append(this_leaf_cluster)
            else:
                leaf_cluster.append(component_node)
            return component_node, node_order

        import abjad
        node_order = []
        node_mapping = {}
        graph = uqbar.graphs.Graph(
            name='G',
            attributes={
                'style': 'rounded',
                },
            edge_attributes={
                },
            node_attributes={
                'fontname': 'Arial',
                'shape': 'none',
                },
            )
        leaf_cluster = uqbar.graphs.Graph(name='leaves')
        component_node, node_order = recurse(self, leaf_cluster)
        if len(leaf_cluster) == 1:
            graph.append(leaf_cluster[0])
        elif len(leaf_cluster):
            graph.append(leaf_cluster)
        graph._node_order = node_order

        if spanner:
            pairs = abjad.sequence(spanner.leaves).nwise()
            for component_one, component_two in pairs:
                node_one = node_mapping[component_one]
                node_two = node_mapping[component_two]
                edge = uqbar.graphs.Edge(
                    attributes={
                        'constraint': False,
                        'penwidth': 5,
                        },
                    )
                edge.attach(node_one, node_two)
            for component in spanner.leaves:
                node = node_mapping[component]
                table = node[0]
                table.attributes['border'] = 4
                table.attributes['bgcolor'] = 'grey80'
                if isinstance(component, Container):
                    for child in abjad.iterate(component)._depth_first():
                        if child is component:
                            continue
                        node = node_mapping[child]
                        table = node[0]
                        table.attributes['bgcolor'] = 'grey80'

        return graph

    def __iter__(self):
        """
        Iterates container.

        ..  container:: example

            Abjad containers are iterables:

            >>> import collections
            >>> container = abjad.Container()
            >>> isinstance(container, collections.Iterable)
            True

        ..  container:: example

            Abjad containers are not sequences:

            >>> import collections
            >>> container = abjad.Container()
            >>> isinstance(container, collections.Sequence)
            False

        Yields container elements.

        Returns generator.
        """
        return iter(self.components)

    def __len__(self) -> int:
        """
        Gets number of components in container.
        """
        return len(self.components)

    def __setitem__(self, i, argument) -> None:
        """
        Sets container ``i`` equal to ``argument``.
        Finds spanners that dominate self[i] and children of self[i].
        Replaces contents at self[i] with 'argument'.
        Reattaches spanners to new contents.
        Always leaves score tree in tact.
        """
        if isinstance(argument, str):
            argument = self._parse_string(argument)
            if isinstance(i, int):
                assert len(argument) == 1, repr(argument)
                argument = argument[0]
        return self._set_item(i, argument)

    ### PRIVATE METHODS ###

    @staticmethod
    def _all_are_orphan_components(argument):
        for component in argument:
            if not isinstance(component, Component):
                return False
            if not inspect(component).parentage().is_orphan:
                return False
        return True

    def _append_without_withdrawing_from_crossing_spanners(self, component):
        """
        Not composer-safe.
        """
        self._set_item(
            slice(len(self), len(self)),
            [component],
            withdraw_components_from_crossing_spanners=False,
            )

    def _as_graphviz_node(self):
        import abjad
        node = Component._as_graphviz_node(self)
        node[0].append(
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    type(self).__name__,
                    attributes={'border': 0},
                    ),
                ])
            )
        return node

    def _copy_with_children(self):
        new_container = self.__copy__()
        for component in self:
            if isinstance(component, Container):
                new_component = component._copy_with_children()
            else:
                new_component = component.__copy__()
            new_container.append(new_component)
        return new_container

    def _eject_contents(self):
        if inspect(self).parentage().parent is not None:
            message = 'can not eject contents of in-score container.'
            raise Exception(message)
        contents = self[:]
        for component in contents:
            component._set_parent(None)
        self._components[:] = []
        return contents

    def _format_after_slot(self, bundle):
        result = []
        result.append(('commands', bundle.after.commands))
        result.append(('comments', bundle.after.comments))
        return tuple(result)

    def _format_before_slot(self, bundle):
        result = []
        result.append(('comments', bundle.before.comments))
        result.append(('commands', bundle.before.commands))
        return tuple(result)

    def _format_close_brackets_slot(self, bundle):
        result = []
        if self.is_simultaneous:
            if self.identifier:
                brackets_close = [f'>>  {self.identifier}']
            else:
                brackets_close = ['>>']
        else:
            if self.identifier:
                brackets_close = [f'}}   {self.identifier}']
            else:
                brackets_close = ['}']
        result.append([('close brackets', ''), brackets_close])
        return tuple(result)

    def _format_closing_slot(self, bundle):
        result = []
        result.append(('grob reverts', bundle.grob_reverts))
        result.append(('commands', bundle.closing.commands))
        result.append(('comments', bundle.closing.comments))
        return self._format_slot_contributions_with_indent(result)

    def _format_content_pieces(self):
        import abjad
        indent = abjad.LilyPondFormatManager.indent
        result = []
        for component in self.components:
            string = component.__format__(
                format_specification='lilypond'
                )
            parts = string.split('\n')
            result.extend(parts)
        result = [indent + _ for _ in result]
        return result

    def _format_contents_slot(self, bundle):
        result = []
        result.append(
            [
                ('contents', '_contents'),
                self._format_content_pieces()
                ])
        return tuple(result)

    def _format_open_brackets_slot(self, bundle):
        result = []
        if self.is_simultaneous:
            if self.identifier:
                brackets_open = [f'<<  {self.identifier}']
            else:
                brackets_open = ['<<']
        else:
            if self.identifier:
                brackets_open = [f'{{   {self.identifier}']
            else:
                brackets_open = ['{']
        result.append([('open brackets', ''), brackets_open])
        return tuple(result)

    def _format_opening_slot(self, bundle):
        result = []
        result.append(('comments', bundle.opening.comments))
        result.append(('commands', bundle.opening.commands))
        result.append(('grob overrides', bundle.grob_overrides))
        result.append(('context settings', bundle.context_settings))
        return self._format_slot_contributions_with_indent(result)

    def _format_slot_contributions_with_indent(self, slot):
        import abjad
        indent = abjad.LilyPondFormatManager.indent
        result = []
        for contributor, contributions in slot:
            result.append(
                (contributor, tuple([indent + x for x in contributions])))
        return tuple(result)

    def _get_abbreviated_string_format(self):
        if 0 < len(self):
            summary = str(len(self))
        else:
            summary = ''
        if self.is_simultaneous:
            open_bracket_string, close_bracket_string = '<<', '>>'
        else:
            open_bracket_string, close_bracket_string = '{', '}'
        name = self.name
        if name is not None:
            name = '-"{}"'.format(name)
        else:
            name = ''
        if hasattr(self, '_lilypond_type'):
            result = '<{}{}{}{}{}>'
            result = result.format(
                self.lilypond_type,
                name,
                open_bracket_string,
                summary,
                close_bracket_string,
                )
        else:
            result = '<{}{}{}{}>'
            result = result.format(
                name,
                open_bracket_string,
                summary,
                close_bracket_string,
                )
        return result

    def _get_compact_representation(self):
        if not self:
            return '{ }'
        return '{{ {} }}'.format(self._get_contents_summary())

    def _get_contents_duration(self):
        if self.is_simultaneous:
            return max(
                [Duration(0)] +
                [x._get_preprolated_duration() for x in self]
                )
        else:
            duration = Duration(0)
            for x in self:
                duration += x._get_preprolated_duration()
            return duration

    def _get_contents_summary(self):
        if 0 < len(self):
            result = []
            for x in self.components:
                if hasattr(x, '_get_compact_representation_with_tie'):
                    result.append(x._get_compact_representation_with_tie())
                elif hasattr(x, '_get_compact_representation'):
                    result.append(x._get_compact_representation())
                else:
                    result.append(str(x))
            return ' '.join(result)
        else:
            return ''

    def _get_duration_in_seconds(self):
        if self.is_simultaneous:
            return max(
                [Duration(0)] +
                [x._get_duration(in_seconds=True) for x in self]
                )
        else:
            duration = Duration(0)
            for leaf in iterate(self).leaves():
                duration += leaf._get_duration(in_seconds=True)
            return duration

    def _get_format_specification(self):
        import abjad
        repr_text = None
        repr_args_values = []
        repr_kwargs_names = self._get_repr_kwargs_names()
        storage_format_args_values = []
        if self:
            repr_args_values.append(self._get_contents_summary())
            lilypond_format = ' '.join(format(x, 'lilypond') for x in self)
            lilypond_format = lilypond_format.replace('\n', ' ')
            lilypond_format = lilypond_format.replace('\t', ' ')
            lilypond_format = lilypond_format.replace('  ', ' ')
            storage_format_args_values.append(lilypond_format)
            if not self[:].are_leaves():
                repr_text = self._get_abbreviated_string_format()
        return abjad.FormatSpecification(
            client=self,
            repr_args_values=repr_args_values,
            repr_kwargs_names=repr_kwargs_names,
            repr_text=repr_text,
            storage_format_args_values=storage_format_args_values,
            #storage_format_kwargs_names=[],
            )

    def _get_preprolated_duration(self):
        return self._get_contents_duration()

    def _get_repr_kwargs_names(self):
        return ['is_simultaneous', 'name']

    def _get_spanners_that_dominate_component_pair(self, left, right):
        """
        Returns spanners that dominant component pair.
        Returns set (spanner, index) pairs.
        ``left`` must be an Abjad component or None.
        ``right`` must be an Abjad component or None.

        If both ``left`` and ``right`` are components,
        then ``left`` and ``right`` must be logical-voice-contiguous.

        This is a version of Selection._get_dominant_spanners().
        This version is useful for finding spanners that dominant
        a zero-length slice between components, as in staff[2:2].
        """
        if left is None or right is None:
            return []
        left_descendants = inspect(left).descendants()
        left_contained = inspect(left_descendants).spanners()
        right_descendants = inspect(right).descendants()
        right_contained = inspect(right_descendants).spanners()
        dominant_spanners = set(left_contained) & set(right_contained)
        dominant_spanners = list(dominant_spanners)
        right_start_offset = inspect(right).timespan().start_offset
        components_after_gap = []
        for component in inspect(right).lineage():
            if inspect(component).timespan().start_offset == right_start_offset:
                components_after_gap.append(component)
        receipt = []
        for spanner in dominant_spanners:
            for component in components_after_gap:
                if component in spanner:
                    index = spanner._index(component)
                    receipt.append((spanner, index))
                    continue
        return receipt

    def _get_spanners_that_dominate_slice(self, start, stop):
        if start == stop:
            if start == 0:
                left = None
            else:
                left = self[start - 1]
            if len(self) <= stop:
                right = None
            else:
                right = self[stop]
            if left is None:
                left = self._get_sibling(-1)
            if right is None:
                right = self._get_sibling(1)
            spanners_receipt = self._get_spanners_that_dominate_component_pair(
                left,
                right,
                )
        else:
            selection = self[start:stop]
            spanners_receipt = selection._get_dominant_spanners()
        assert isinstance(spanners_receipt, list), repr(spanners_receipt)
        return spanners_receipt

    def _initialize_components(self, components):
        if (isinstance(components, collections.Iterable) and
            not isinstance(components, str)):
            components_ = []
            for item in components:
                if isinstance(item, Selection):
                    components_.extend(item)
                elif isinstance(item, str):
                    parsed = self._parse_string(item)
                    components_.append(parsed)
                else:
                    components_.append(item)
            components = components_
            for component in components:
                if not isinstance(component, Component):
                    message = f'must be component: {component!r}.'
                    raise Exception(component)
        if self._all_are_orphan_components(components):
            self._components = list(components)
            self[:]._set_parents(self)
        elif isinstance(components, str):
            parsed = self._parse_string(components)
            self._components = []
            self.is_simultaneous = parsed.is_simultaneous
            if (
                parsed.is_simultaneous or
                not select(parsed[:]).are_contiguous_logical_voice()
                ):
                while len(parsed):
                    self.append(parsed.pop(0))
            else:
                self[:] = parsed[:]
        else:
            message = 'can not initialize container from {!r}.'
            message = message.format(components)
            raise TypeError(message)

    def _is_one_of_my_first_leaves(self, leaf):
        return leaf in self._get_descendants_starting_with()

    def _is_one_of_my_last_leaves(self, leaf):
        return leaf in self._get_descendants_stopping_with()

    def _iterate_bottom_up(self):
        def recurse(node):
            if isinstance(node, Container):
                for x in node:
                    for y in recurse(x):
                        yield y
            yield node
        return recurse(self)

    def _iterate_top_down(self):
        def recurse(node):
            yield node
            if isinstance(node, Container):
                for x in node:
                    for y in recurse(x):
                        yield y
        return recurse(self)

    def _iterate_topmost(self):
        import abjad
        for component in self:
            if isinstance(component, abjad.Leaf):
                ties = inspect(component).spanners(abjad.Tie)
                if not ties or tuple(ties)[0].leaves[-1] is component:
                    yield inspect(component).logical_tie()
            else:
                assert isinstance(component, abjad.Container)
                yield component

    def _move_spanners_to_children(self):
        for spanner in inspect(self).spanners():
            i = spanner._index(self)
            spanner._components.__setitem__(slice(i, i + 1), self[:])
            for component in self:
                component._append_spanner(spanner)
            self._remove_spanner(spanner)
        return self

    def _parse_string(self, string):
        import abjad
        user_input = string.strip()
        if user_input.startswith('abj:'):
            parser = abjad.parser.ReducedLyParser()
            parsed = parser(user_input[4:])
            if parser._toplevel_component_count == 1:
                parent = inspect(parsed).parentage().parent
                if parent is None:
                    parsed = Container([parsed])
                else:
                    parsed = parent
        elif user_input.startswith('rtm:'):
            parsed = abjad.rhythmtrees.parse_rtm_syntax(user_input[4:])
        else:
            if (
                not user_input.startswith('<<') or
                not user_input.endswith('>>')
                ):
                user_input = '{{ {} }}'.format(user_input)
            parsed = abjad.parse(user_input)
            if isinstance(parsed, abjad.LilyPondFile):
                parsed = Container(parsed.items[:])
            assert isinstance(parsed, Container)
        return parsed

    @staticmethod
    def _remove_powers_of_two(n):
        assert isinstance(n, int), repr(n)
        assert not n <= 0, repr(n)
        while n % 2 == 0:
            n //= 2
        return n

    def _scale(self, multiplier):
        self._scale_contents(multiplier)

    def _scale_contents(self, multiplier):
        for item in  list(self._iterate_topmost()):
            item._scale(multiplier)

    def _set_item(
        self,
        i,
        argument,
        withdraw_components_from_crossing_spanners=True,
        ):
        """
        This method exists because __setitem__ can not accept keywords.

        Note that setting
        withdraw_components_from_crossing_spanners=False constitutes a
        composer-unsafe use of this method.

        Only private methods should set this keyword.
        """
        import abjad
        argument_indicators = []
        for component in abjad.iterate(argument).components():
            wrappers = inspect(component).wrappers()
            argument_indicators.extend(wrappers)
        if isinstance(i, int):
            argument = [argument]
            if i < 0:
                i = len(self) + i
            i = slice(i, i + 1)
        prototype = (Component, Selection)
        assert all(isinstance(_, prototype) for _ in argument)
        new_argument = []
        for item in argument:
            if isinstance(item, Selection):
                new_argument.extend(item)
            else:
                new_argument.append(item)
        argument = new_argument
        assert all(isinstance(_, abjad.Component) for _ in argument)
        if any(isinstance(_, abjad.GraceContainer) for _ in argument):
            message = 'must attach grace container to note or chord.'
            raise Exception(message)
        if self._check_for_cycles(argument):
            raise exceptions.ParentageError('attempted to induce cycles.')
        if (i.start == i.stop and
            i.start is not None and
            i.stop is not None and
            i.start <= -len(self)):
            start, stop = 0, 0
        else:
            start, stop, stride = i.indices(len(self))
        old_components = self[start:stop]
        spanners_receipt = self._get_spanners_that_dominate_slice(start, stop)
        for component in old_components:
            for child in abjad.iterate([component]).components():
                for spanner in inspect(child).spanners():
                    spanner._remove(child)
        del(self[start:stop])
        # must withdraw before setting in self!
        # otherwise circular withdraw ensues!
        if withdraw_components_from_crossing_spanners:
            selection = select(argument)
            if selection.are_contiguous_logical_voice():
                selection._withdraw_from_crossing_spanners()
        self._components.__setitem__(slice(start, start), argument)
        for component in argument:
            component._set_parent(self)
        for spanner, index in spanners_receipt:
            for component in reversed(argument):
                # attach spanners only to leaves
                leaves = select(component).leaves()
                for leaf in reversed(leaves):
                    spanner._insert(index, leaf)
                    leaf._append_spanner(spanner)
        for indicator in argument_indicators:
            if hasattr(indicator, '_update_effective_context'):
                indicator._update_effective_context()

    def _split_at_index(self, i, fracture_spanners=False):
        """
        Splits container to the left of index ``i``.

        Preserves tuplet multiplier when container is a tuplet.

        Preserves time signature denominator when container is a measure.

        Resizes resizable containers.

        Returns split parts.
        """
        import abjad
        # partition my components
        left_components = self[:i]
        right_components = self[i:]
        # instantiate new left and right containers
        if isinstance(self, abjad.Measure):
            time_signature = self._get_effective(abjad.TimeSignature)
            denominator = time_signature.denominator
            left_duration = sum([_._get_duration() for _ in left_components])
            left_pair = abjad.NonreducedFraction(left_duration)
            left_pair = left_pair.with_multiple_of_denominator(denominator)
            left_time_signature = abjad.TimeSignature(left_pair)
            left = type(self)(left_time_signature, [])
            abjad.mutate(left_components).wrap(left)
            left.implicit_scaling = self.implicit_scaling
            right_duration = sum([_._get_duration() for _ in right_components])
            right_pair = abjad.NonreducedFraction(right_duration)
            right_pair = right_pair.with_multiple_of_denominator(denominator)
            right_time_signature = abjad.TimeSignature(right_pair)
            right = type(self)(right_time_signature, [])
            abjad.mutate(right_components).wrap(right)
            right.implicit_scaling = self.implicit_scaling
        elif isinstance(self, abjad.Tuplet):
            multiplier = self.multiplier
            left = type(self)(multiplier, [])
            abjad.mutate(left_components).wrap(left)
            right = type(self)(multiplier, [])
            abjad.mutate(right_components).wrap(right)
        else:
            left = self.__copy__()
            abjad.mutate(left_components).wrap(left)
            right = self.__copy__()
            abjad.mutate(right_components).wrap(right)
        # save left and right containers together for iteration
        halves = (left, right)
        nonempty_halves = [half for half in halves if len(half)]
        # give my attached spanners to my children
        self._move_spanners_to_children()
        # incorporate left and right parents in score if possible
        selection = select(self)
        parent, start, stop = selection._get_parent_and_start_stop_indices()
        if parent is not None:
            parent._components.__setitem__(slice(start, stop + 1), nonempty_halves)
            for part in nonempty_halves:
                part._set_parent(parent)
        else:
            left._set_parent(None)
            right._set_parent(None)
        # fracture spanners if requested
        if fracture_spanners:
            for spanner in inspect(left).spanners():
                index = spanner._index(left)
                spanner._fracture(index, direction=enums.Right)
        # return new left and right containers
        return halves

    def _split_by_duration(
        self,
        duration,
        fracture_spanners=False,
        tie_split_notes=True,
        repeat_ties=False,
        ):
        import abjad
        if self.is_simultaneous:
            return self._split_simultaneous_by_duration(
                duration=duration,
                fracture_spanners=fracture_spanners,
                tie_split_notes=tie_split_notes,
                repeat_ties=repeat_ties,
                )
        # check input
        duration = Duration(duration)
        assert 0 <= duration, repr(duration)
        # if zero duration then return empty list and self
        if duration == 0:
            return [], self
        # get split point score offset
        timespan = inspect(self).timespan()
        global_split_point = timespan.start_offset + duration
        # get any duration-crossing descendents
        cross_offset = timespan.start_offset + duration
        duration_crossing_descendants = []
        for descendant in inspect(self).descendants():
            timespan = inspect(descendant).timespan()
            start_offset = timespan.start_offset
            stop_offset = timespan.stop_offset
            if start_offset < cross_offset < stop_offset:
                duration_crossing_descendants.append(descendant)
        # get any duration-crossing measure descendents
        measures = [
            _ for _ in duration_crossing_descendants
            if isinstance(_, abjad.Measure)
            ]
        # if we must split a power-of-two measure at non-power-of-two
        # split point then go ahead and transform the power-of-two measure
        # to non-power-of-two equivalent now;
        # code that crawls and splits later on will be happier
        if len(measures) == 1:
            measure = measures[0]
            timespan = inspect(measure).timespan()
            start_offset = timespan.start_offset
            split_point_in_measure = global_split_point - start_offset
            if measure.has_non_power_of_two_denominator:
                pass
            elif not abjad.mathtools.is_nonnegative_integer_power_of_two(
                split_point_in_measure.denominator):
                non_power_of_two_factors = self._remove_powers_of_two(
                    split_point_in_measure.denominator
                    )
                non_power_of_two_factors = abjad.mathtools.factors(
                    non_power_of_two_factors)
                non_power_of_two_product = 1
                for non_power_of_two_factor in non_power_of_two_factors:
                    non_power_of_two_product *= non_power_of_two_factor
                measure._scale_denominator(non_power_of_two_product)
                # rederive duration crossers with possibly new measure contents
                timespan = inspect(self).timespan()
                cross_offset = timespan.start_offset + duration
                duration_crossing_descendants = []
                for descendant in inspect(self).descendants():
                    timespan = inspect(descendant).timespan()
                    start_offset = timespan.start_offset
                    stop_offset = timespan.stop_offset
                    if start_offset < cross_offset < stop_offset:
                        duration_crossing_descendants.append(descendant)
        elif 1 < len(measures):
            message = 'measures can not nest.'
            raise Exception(message)
        # any duration-crossing leaf will be at end of list
        bottom = duration_crossing_descendants[-1]
        did_split_leaf = False
        # if split point necessitates leaf split
        if isinstance(bottom, abjad.Leaf):
            assert isinstance(bottom, abjad.Leaf)
            did_split_leaf = True
            timespan = inspect(bottom).timespan()
            start_offset = timespan.start_offset
            split_point_in_bottom = global_split_point - start_offset
            left_list, right_list = bottom._split_by_durations(
                [split_point_in_bottom],
                fracture_spanners=fracture_spanners,
                tie_split_notes=tie_split_notes,
                repeat_ties=repeat_ties,
                )
            right = right_list[0]
            leaf_right_of_split = right
            leaf_left_of_split = left_list[-1]
            duration_crossing_containers = duration_crossing_descendants[:-1]
            if not len(duration_crossing_containers):
                return left_list, right_list
        # if split point falls between leaves
        # then find leaf to immediate right of split point
        # in order to start upward crawl through duration-crossing containers
        else:
            duration_crossing_containers = duration_crossing_descendants[:]
            for leaf in abjad.iterate(bottom).leaves():
                timespan = inspect(leaf).timespan()
                if timespan.start_offset == global_split_point:
                    leaf_right_of_split = leaf
                    leaf_left_of_split = inspect(leaf).leaf(-1)
                    break
            else:
                message = 'can not split empty container {!r}.'
                message = message.format(bottom)
                raise Exception(message)
        assert leaf_left_of_split is not None
        assert leaf_right_of_split is not None
        # find component to right of split
        # that is also immediate child of last duration-crossing container
        agent = inspect(leaf_right_of_split)
        parentage = agent.parentage(include_self=True)
        for component in parentage:
            parent = inspect(component).parentage().parent
            if parent is duration_crossing_containers[-1]:
                highest_level_component_right_of_split = component
                break
        else:
            message = 'should not be able to get here.'
            raise ValueError(message)
        # crawl back up through duration-crossing containers
        # and fracture spanners if requested
        if fracture_spanners:
            agent = inspect(leaf_right_of_split)
            start_offset = agent.timespan().start_offset
            for parent in agent.parentage():
                timespan = inspect(parent).timespan()
                if timespan.start_offset == start_offset:
                    for spanner in inspect(parent).spanners():
                        index = spanner._index(parent)
                        spanner._fracture(index, direction=enums.Left)
                if parent is component:
                    break
        # crawl back up through duration-crossing containers and split each
        previous = highest_level_component_right_of_split
        for container in reversed(duration_crossing_containers):
            assert isinstance(container, abjad.Container)
            index = container.index(previous)
            left, right = container._split_at_index(
                index,
                fracture_spanners=fracture_spanners,
                )
            previous = right
        # NOTE: If logical tie here is convenience, then fusing is good.
        #       If logical tie here is user-given, then fusing is less good.
        #       Maybe later model difference between user logical ties and not.
        left_logical_tie = inspect(leaf_left_of_split).logical_tie()
        right_logical_tie = inspect(leaf_right_of_split).logical_tie()
        left_logical_tie._fuse_leaves_by_immediate_parent()
        right_logical_tie._fuse_leaves_by_immediate_parent()
        # reapply tie here if crawl above killed tie applied to leaves
        if did_split_leaf:
            if (
                tie_split_notes and
                isinstance(leaf_left_of_split, abjad.Note)
                ):
                if (
                    inspect(leaf_left_of_split).parentage().root is
                    inspect(leaf_right_of_split).parentage().root
                    ):
                    leaves_around_split = (
                        leaf_left_of_split,
                        leaf_right_of_split,
                        )
                    selection = select(leaves_around_split)
                    selection._attach_tie_to_leaf_pair(
                        repeat_ties=repeat_ties,
                        )
        # return list-wrapped halves of container
        return [left], [right]

    def _split_simultaneous_by_duration(
        self,
        duration,
        fracture_spanners=False,
        tie_split_notes=True,
        repeat_ties=False,
        ):
        import abjad
        assert self.is_simultaneous
        left_components, right_components = [], []
        for component in self[:]:
            halves = component._split_by_duration(
                duration=duration,
                fracture_spanners=fracture_spanners,
                tie_split_notes=tie_split_notes,
                repeat_ties=repeat_ties,
                )
            left_components_, right_components_ = halves
            left_components.extend(left_components_)
            right_components.extend(right_components_)
        left_components = select(left_components)
        right_components = select(right_components)
        left_container = self.__copy__()
        right_container = self.__copy__()
        left_container.extend(left_components)
        right_container.extend(right_components)
        if inspect(self).parentage().parent is not None:
            containers = select([left_container, right_container])
            abjad.mutate(self).replace(containers)
        # return list-wrapped halves of container
        return [left_container], [right_container]

    ### PUBLIC PROPERTIES ###

    @property
    def components(self) -> tuple:
        """
        Gets components in container.
        """
        return self._components

    @property
    def identifier(self)  -> typing.Optional[str]:
        r"""
        Gets and sets bracket comment.

        ..  container:: example

            >>> container = abjad.Container(
            ...     "c'4 d'4 e'4 f'4",
            ...     identifier='%*% AB',
            ...     )
            >>> abjad.show(container) # doctest: +SKIP

            >>> abjad.f(container)
            {   %*% AB
                c'4
                d'4
                e'4
                f'4
            }   %*% AB

        """
        return self._identifier

    @identifier.setter
    def identifier(self, argument):
        assert isinstance(argument, (str, type(None))), repr(argument)
        self._identifier: typing.Optional[str] = argument

    @property
    def is_simultaneous(self) -> typing.Optional[bool]:
        r"""
        Is true when container is simultaneous.

        ..  container:: example

            Gets simultaneity status of container:

            >>> container = abjad.Container()
            >>> container.append(abjad.Voice("c'8 d'8 e'8"))
            >>> container.append(abjad.Voice('g4.'))
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    \new Voice
                    {
                        c'8
                        d'8
                        e'8
                    }
                    \new Voice
                    {
                        g4.
                    }
                }

            >>> container.is_simultaneous is None
            True

        ..  container:: example

            Sets simultaneity status of container:

            >>> container = abjad.Container()
            >>> container.append(abjad.Voice("c'8 d'8 e'8"))
            >>> container.append(abjad.Voice('g4.'))
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    \new Voice
                    {
                        c'8
                        d'8
                        e'8
                    }
                    \new Voice
                    {
                        g4.
                    }
                }

            >>> container.is_simultaneous = True
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                <<
                    \new Voice
                    {
                        c'8
                        d'8
                        e'8
                    }
                    \new Voice
                    {
                        g4.
                    }
                >>

        """
        return self._is_simultaneous

    @is_simultaneous.setter
    def is_simultaneous(self, argument):
        if argument is None:
            return
        assert isinstance(argument, bool), repr(argument)
        if argument and not all(isinstance(_, Container) for _ in self):
            message = 'simultaneous containers must contain'
            message += ' only other containers.'
            raise ValueError(message)
        self._is_simultaneous = argument
        self._update_later(offsets=True)

    @property
    def name(self) -> typing.Optional[str]:
        r"""
        Gets and sets name of container.

        ..  container:: example

            Gets container name:

            >>> container = abjad.Container("c'4 d'4 e'4 f'4")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }

            >>> container.name is None
            True

        ..  container:: example

            Sets container name:

            >>> container = abjad.Container(
            ...     "c'4 d'4 e'4 f'4",
            ...     name='Special',
            ...     )
            >>> abjad.show(container) # doctest: +SKIP

            >>> container.name
            'Special'

            Container name does not appear in LilyPond output:

            >>> abjad.f(container)
            {
                c'4
                d'4
                e'4
                f'4
            }

        """
        return self._name

    @name.setter
    def name(self, argument):
        assert isinstance(argument, (str, type(None)))
        old_name = self._name
        for parent in inspect(self).parentage(include_self=False):
            named_children = parent._named_children
            if old_name is not None:
                named_children[old_name].remove(self)
                if not named_children[old_name]:
                    del named_children[old_name]
            if argument is not None:
                if argument not in named_children:
                    named_children[argument] = [self]
                else:
                    named_children[argument].append(self)
        self._name = argument

    ### PUBLIC METHODS ###

    def append(self, component) -> None:
        r"""
        Appends ``component`` to container.

        ..  container:: example

            Appends note to container:

            >>> container = abjad.Container("c'4 ( d'4 f'4 )")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> container.append(abjad.Note("e'4"))
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'4
                }

        """
        if isinstance(component, str):
            selection = self._parse_string(component)
            assert len(selection) == 1
            component = selection[0]
        self.__setitem__(slice(len(self), len(self)), [component])

    def extend(self, argument) -> None:
        r"""
        Extends container with ``argument``.

        ..  container:: example

            Extends container with three notes:

            >>> container = abjad.Container("c'4 ( d'4 f'4 )")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> notes = [abjad.Note("e'32"), abjad.Note("d'32"), abjad.Note("e'16")]
            >>> container.extend(notes)
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'32
                    d'32
                    e'16
                }

        """
        if isinstance(argument, str):
            argument = self._parse_string(argument)
        elif isinstance(argument, collections.Iterable):
            argument_ = []
            for item in argument:
                if isinstance(item, str):
                    item = self._parse_string(item)
                argument_.append(item)
            argument = argument_
        self.__setitem__(
            slice(len(self), len(self)),
            argument.__getitem__(slice(0, len(argument)))
            )

    def index(self, component) -> int:
        r"""
        Returns index of ``component`` in container.

        ..  container:: example

            Gets index of last element in container:

            >>> container = abjad.Container("c'4 d'4 f'4 e'4")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    d'4
                    f'4
                    e'4
                }

            >>> note = container[-1]
            >>> note
            Note("e'4")

            >>> container.index(note)
            3

        """
        for i, element in enumerate(self.components):
            if element is component:
                return i
        else:
            message = 'component {!r} not in Abjad container {!r}.'
            message = message.format(component, self)
            raise ValueError(message)

    def insert(self, i, component, fracture_spanners=False) -> None:
        r"""
        Inserts ``component`` at index ``i`` in container.

        ..  container:: example

            Inserts note. Does not fracture spanners:

            >>> container = abjad.Container([])
            >>> container.extend("fs16 cs' e' a'")
            >>> container.extend("cs''16 e'' cs'' a'")
            >>> container.extend("fs'16 e' cs' fs")
            >>> slur = abjad.Slur(direction=abjad.Down)
            >>> abjad.attach(slur, container[:])
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    fs16
                    _ (
                    cs'16
                    e'16
                    a'16
                    cs''16
                    e''16
                    cs''16
                    a'16
                    fs'16
                    e'16
                    cs'16
                    fs16
                    )
                }

            >>> container.insert(-4, abjad.Note("e'4"), fracture_spanners=False)
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    fs16
                    _ (
                    cs'16
                    e'16
                    a'16
                    cs''16
                    e''16
                    cs''16
                    a'16
                    e'4
                    fs'16
                    e'16
                    cs'16
                    fs16
                    )
                }

        ..  container:: example

            Inserts note. Fractures spanners:

            >>> container = abjad.Container([])
            >>> container.extend("fs16 cs' e' a'")
            >>> container.extend("cs''16 e'' cs'' a'")
            >>> container.extend("fs'16 e' cs' fs")
            >>> slur = abjad.Slur(direction=abjad.Down)
            >>> abjad.attach(slur, container[:])
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    fs16
                    _ (
                    cs'16
                    e'16
                    a'16
                    cs''16
                    e''16
                    cs''16
                    a'16
                    fs'16
                    e'16
                    cs'16
                    fs16
                    )
                }

            >>> container.insert(-4, abjad.Note("e'4"), fracture_spanners=True)
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    fs16
                    _ (
                    cs'16
                    e'16
                    a'16
                    cs''16
                    e''16
                    cs''16
                    a'16
                    )
                    e'4
                    fs'16
                    _ (
                    e'16
                    cs'16
                    fs16
                    )
                }

        """
        assert isinstance(i, int)
        if isinstance(component, str):
            selection = self._parse_string(component)
            assert len(selection) == 1, repr(selection)
            component = selection[0]
        if not fracture_spanners:
            self.__setitem__(slice(i, i), [component])
            return
        assert isinstance(component, Component)
        component._set_parent(self)
        self._components.insert(i, component)
        previous_leaf = component._get_leaf(-1)
        if previous_leaf:
            for spanner in inspect(previous_leaf).spanners():
                index = spanner._index(previous_leaf)
                spanner._fracture(index, direction=enums.Right)
        next_leaf = component._get_leaf(1)
        if next_leaf:
            for spanner in inspect(next_leaf).spanners():
                index = spanner._index(next_leaf)
                spanner._fracture(index, direction=enums.Left)

    def pop(self, i=-1):
        r"""
        Pops component from container at index ``i``.

        ..  container:: example

            Pops last element from container:

            >>> container = abjad.Container("c'4 ( d'4 f'4 ) e'4")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'4
                }

            >>> container.pop()
            Note("e'4")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

        Returns component.
        """
        component = self[i]
        del(self[i])
        return component

    def remove(self, component) -> None:
        r"""
        Removes ``component`` from container.

        ..  container:: example

            Removes note from container:

            >>> container = abjad.Container("c'4 ( d'4 f'4 ) e'4")
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'4
                }

            >>> note = container[2]
            >>> note
            Note("f'4")

            >>> container.remove(note)
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    c'4
                    (
                    d'4
                    )
                    e'4
                }

        """
        i = self.index(component)
        del(self[i])



from .NoteHead import NoteHead


class DrumNoteHead(NoteHead):
    """
    Drum note-head.

    ..  container:: example

        >>> note_head = abjad.DrumNoteHead('snare')
        >>> note_head
        DrumNoteHead('snare')

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Note-heads'

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(
        self,
        written_pitch='snare',
        client=None,
        is_cautionary=None,
        is_forced=None,
        is_parenthesized=None,
        tweaks=(),
        ):
        from abjad.ly import drums
        NoteHead.__init__(
            self,
            written_pitch=None,
            client=client,
            is_cautionary=is_cautionary,
            is_forced=is_forced,
            is_parenthesized=is_parenthesized,
            tweaks=tweaks,
            )
        assert str(written_pitch) in drums
        drum_pitch = drums[str(written_pitch)]
        self._written_pitch = drum_pitch



import collections
import numbers
from abjad.system.AbjadValueObject import AbjadValueObject


class LeafMaker(AbjadValueObject):
    r"""
    Leaf-maker.

    ..  container:: example

        Integer and string elements in ``pitches`` result in notes:

        >>> maker = abjad.LeafMaker()
        >>> pitches = [2, 4, 'F#5', 'G#5']
        >>> duration = abjad.Duration(1, 4)
        >>> leaves = maker(pitches, duration)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                d'4
                e'4
                fs''4
                gs''4
            }

    ..  container:: example

        Tuple elements in ``pitches`` result in chords:

        >>> maker = abjad.LeafMaker()
        >>> pitches = [(0, 2, 4), ('F#5', 'G#5', 'A#5')]
        >>> duration = abjad.Duration(1, 2)
        >>> leaves = maker(pitches, duration)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                <c' d' e'>2
                <fs'' gs'' as''>2
            }

    ..  container:: example

        None-valued elements in ``pitches`` result in rests:

        >>> maker = abjad.LeafMaker()
        >>> pitches = 4 * [None]
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> staff.lilypond_type = 'RhythmicStaff'
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new RhythmicStaff
            {
                r4
                r4
                r4
                r4
            }

    ..  container:: example

        You can mix and match values passed to ``pitches``:

        >>> maker = abjad.LeafMaker()
        >>> pitches = [(0, 2, 4), None, 'C#5', 'D#5']
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                <c' d' e'>4
                r4
                cs''4
                ds''4
            }

    ..  container:: example

        Works with segments:

        >>> maker = abjad.LeafMaker()
        >>> pitches = abjad.PitchSegment("e'' ef'' d'' df'' c''")
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                e''4
                ef''4
                d''4
                df''4
                c''4
            }

    ..  container:: example

        Reads ``pitches`` cyclically when the length of ``pitches`` is less
        than the length of ``durations``:

        >>> maker = abjad.LeafMaker()
        >>> pitches = ['C5']
        >>> durations = 2 * [abjad.Duration(3, 8), abjad.Duration(1, 8)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c''4.
                c''8
                c''4.
                c''8
            }

    ..  container:: example

        Reads ``durations`` cyclically when the length of ``durations`` is less
        than the length of ``pitches``:

        >>> maker = abjad.LeafMaker()
        >>> pitches = "c'' d'' e'' f''"
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c''4
                d''4
                e''4
                f''4
            }

    ..  container:: example

        Elements in ``durations`` with non-power-of-two denominators result in
        tuplet-nested leaves:

        >>> maker = abjad.LeafMaker()
        >>> pitches = ['D5']
        >>> durations = 3 * [abjad.Duration(1, 3)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \times 2/3 {
                    d''2
                    d''2
                    d''2
                }
            }

    ..  container:: example

        Set ``decrease_monotonic`` to true to return nonassignable
        durations tied from greatest to least:

        >>> maker = abjad.LeafMaker()
        >>> pitches = ['D#5']
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> time_signature = abjad.TimeSignature((13, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \time 13/16
                ds''2.
                ~
                ds''16
            }

    ..  container:: example

        Set ``decrease_monotonic`` to false to return nonassignable
        durations tied from least to greatest:

        >>> maker = abjad.LeafMaker(decrease_monotonic=False)
        >>> pitches = ['E5']
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> time_signature = abjad.TimeSignature((13, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \time 13/16
                e''16
                ~
                e''2.
            }

    ..  container:: example

        Set ``forbidden_duration`` to avoid notes greater than or equal
        to a certain written duration:

        >>> maker = abjad.LeafMaker(
        ...     forbidden_duration=abjad.Duration(1, 2),
        ...     )
        >>> pitches = "f' g'"
        >>> durations = [abjad.Duration(5, 8)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> time_signature = abjad.TimeSignature((5, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \time 5/4
                f'4
                ~
                f'4
                ~
                f'8
                g'4
                ~
                g'4
                ~
                g'8
            }

    ..  container:: example

        You may set ``forbidden_duration`` and ``decrease_monotonic`` together:

        >>> maker = abjad.LeafMaker(
        ...     forbidden_duration=abjad.Duration(1, 2),
        ...     decrease_monotonic=False,
        ...     )
        >>> pitches = "f' g'"
        >>> durations = [abjad.Duration(5, 8)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> time_signature = abjad.TimeSignature((5, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \time 5/4
                f'8
                ~
                f'4
                ~
                f'4
                g'8
                ~
                g'4
                ~
                g'4
            }

    ..  container:: example

        Produces diminished tuplets:

        >>> maker = abjad.LeafMaker()
        >>> pitches = "f'"
        >>> durations = [abjad.Duration(5, 14)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> time_signature = abjad.TimeSignature((5, 14))
        >>> leaf = abjad.inspect(staff).leaf(0)
        >>> abjad.attach(time_signature, leaf)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \tweak edge-height #'(0.7 . 0)
                \times 4/7 {
                    \time 5/14
                    f'2
                    ~
                    f'8
                }
            }

        This is default behavior.

    ..  container:: example

        None-valued elements in ``pitches`` result in multimeasure rests when
        the multimeasure rest keyword is set:

        >>> maker = abjad.LeafMaker(use_multimeasure_rests=True)
        >>> pitches = [None]
        >>> durations = [abjad.Duration(3, 8), abjad.Duration(5, 8)]
        >>> leaves = maker(pitches, durations)
        >>> leaves
        Selection([MultimeasureRest('R1 * 3/8'), MultimeasureRest('R1 * 5/8')])

        >>> staff = abjad.Staff([
        ...     abjad.Measure((3, 8), [leaves[0]]),
        ...     abjad.Measure((5, 8), [leaves[1]]),
        ...     ])
        >>> staff.lilypond_type = 'RhythmicStaff'
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new RhythmicStaff
            {
                {   % measure
                    \time 3/8
                    R1 * 3/8
                }   % measure
                {   % measure
                    \time 5/8
                    R1 * 5/8
                }   % measure
            }

    ..  container:: example

        Uses repeat ties:

        >>> maker = abjad.LeafMaker(repeat_ties=True)
        >>> pitches = [0]
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'2.
                c'16
                \repeatTie
            }

    ..  container:: example

        Works with numbered pitch-class:

        >>> maker = abjad.LeafMaker()
        >>> pitches = [abjad.NumberedPitchClass(6)]
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = maker(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                fs'2.
                ~
                fs'16
            }

    ..  container:: example

        Makes skips instead of rests:

        >>> maker = abjad.LeafMaker(skips_instead_of_rests=True)
        >>> pitches = [None]
        >>> durations = [abjad.Duration(13, 16)]
        >>> maker(pitches, durations)
        Selection([Skip('s2.'), Skip('s16')])

    Returns selection of leaves.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Makers'

    __slots__ = (
        '_decrease_monotonic',
        '_forbidden_duration',
        '_metrical_hierarchy',
        '_skips_instead_of_rests',
        '_repeat_ties',
        '_use_multimeasure_rests',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        decrease_monotonic=True,
        forbidden_duration=None,
        metrical_hierarchy=None,
        skips_instead_of_rests=False,
        repeat_ties=False,
        use_multimeasure_rests=False,
        ):
        self._decrease_monotonic = decrease_monotonic
        self._forbidden_duration = forbidden_duration
        self._metrical_hierarchy = metrical_hierarchy
        self._skips_instead_of_rests = skips_instead_of_rests
        self._repeat_ties = repeat_ties
        self._use_multimeasure_rests = use_multimeasure_rests

    ### SPECIAL METHODS ###

    def __call__(self, pitches, durations):
        """
        Calls leaf-maker on ``pitches`` and ``durations``.

        Returns selection.
        """
        import abjad
        if isinstance(pitches, str):
            pitches = pitches.split()
        if not isinstance(pitches, collections.Iterable):
            pitches = [pitches]
        if isinstance(durations, (numbers.Number, tuple)):
            durations = [durations]
        nonreduced_fractions = [abjad.NonreducedFraction(_) for _ in durations]
        size = max(len(nonreduced_fractions), len(pitches))
        nonreduced_fractions = abjad.sequence(nonreduced_fractions)
        nonreduced_fractions = nonreduced_fractions.repeat_to_length(size)
        pitches = abjad.sequence(pitches).repeat_to_length(size)
        Duration = abjad.Duration
        duration_groups = Duration._group_by_implied_prolation(
            nonreduced_fractions
            )
        result = []
        for duration_group in duration_groups:
            # get factors in denominator of duration group other than 1, 2.
            factors = abjad.mathtools.factors(duration_group[0].denominator)
            factors = set(factors)
            factors.discard(1)
            factors.discard(2)
            current_pitches = pitches[0:len(duration_group)]
            pitches = pitches[len(duration_group):]
            if len(factors) == 0:
                for pitch, duration in zip(current_pitches, duration_group):
                    leaves = self._make_leaf_on_pitch(
                        pitch,
                        duration,
                        decrease_monotonic=self.decrease_monotonic,
                        forbidden_duration=self.forbidden_duration,
                        skips_instead_of_rests=self.skips_instead_of_rests,
                        use_multimeasure_rests=self.use_multimeasure_rests,
                        repeat_ties=self.repeat_ties,
                        )
                    result.extend(leaves)
            else:
                # compute tuplet prolation
                denominator = duration_group[0].denominator
                numerator = abjad.mathtools.greatest_power_of_two_less_equal(
                    denominator)
                multiplier = (numerator, denominator)
                ratio = 1 / abjad.Duration(*multiplier)
                duration_group = [
                    ratio * abjad.Duration(duration)
                    for duration in duration_group
                    ]
                # make tuplet leaves
                tuplet_leaves = []
                for pitch, duration in zip(current_pitches, duration_group):
                    leaves = self._make_leaf_on_pitch(
                        pitch,
                        duration,
                        decrease_monotonic=self.decrease_monotonic,
                        skips_instead_of_rests=self.skips_instead_of_rests,
                        use_multimeasure_rests=self.use_multimeasure_rests,
                        repeat_ties=self.repeat_ties,
                        )
                    tuplet_leaves.extend(leaves)
                tuplet = abjad.Tuplet(multiplier, tuplet_leaves)
                result.append(tuplet)
        return abjad.select(result)

    ### PRIVATE METHODS ###

    @staticmethod
    def _make_leaf_on_pitch(
        pitch,
        duration,
        decrease_monotonic=True,
        forbidden_duration=None,
        skips_instead_of_rests=False,
        use_multimeasure_rests=False,
        repeat_ties=False,
        ):
        import abjad
        note_prototype = (
            numbers.Number,
            str,
            abjad.NamedPitch,
            abjad.NumberedPitch,
            abjad.PitchClass,
            )
        chord_prototype = (tuple, list)
        rest_prototype = (type(None),)
        if isinstance(pitch, note_prototype):
            leaves = LeafMaker._make_tied_leaf(
                abjad.Note,
                duration,
                decrease_monotonic=decrease_monotonic,
                forbidden_duration=forbidden_duration,
                pitches=pitch,
                repeat_ties=repeat_ties,
                )
        elif isinstance(pitch, chord_prototype):
            leaves = LeafMaker._make_tied_leaf(
                abjad.Chord,
                duration,
                decrease_monotonic=decrease_monotonic,
                forbidden_duration=forbidden_duration,
                pitches=pitch,
                repeat_ties=repeat_ties,
                )
        elif isinstance(pitch, rest_prototype) and skips_instead_of_rests:
            leaves = LeafMaker._make_tied_leaf(
                abjad.Skip,
                duration,
                decrease_monotonic=decrease_monotonic,
                forbidden_duration=forbidden_duration,
                pitches=None,
                repeat_ties=repeat_ties,
                )
        elif isinstance(pitch, rest_prototype) and not use_multimeasure_rests:
            leaves = LeafMaker._make_tied_leaf(
                abjad.Rest,
                duration,
                decrease_monotonic=decrease_monotonic,
                forbidden_duration=forbidden_duration,
                pitches=None,
                repeat_ties=repeat_ties,
                )
        elif isinstance(pitch, rest_prototype) and use_multimeasure_rests:
            multimeasure_rest = abjad.MultimeasureRest((1))
            multiplier = abjad.Multiplier(duration)
            abjad.attach(multiplier, multimeasure_rest)
            leaves = (
                multimeasure_rest,
                )
        else:
            message = 'unknown pitch: {!r}.'
            message = message.format(pitch)
            raise ValueError(message)
        return leaves

    @staticmethod
    def _make_tied_leaf(
        class_,
        duration,
        decrease_monotonic=True,
        forbidden_duration=None,
        pitches=None,
        tie_parts=True,
        repeat_ties=False,
        ):
        import abjad
        # check input
        duration = abjad.Duration(duration)
        if forbidden_duration is not None:
            forbidden_duration = abjad.Duration(forbidden_duration)
            assert forbidden_duration.is_assignable
            assert forbidden_duration.numerator == 1
        # find preferred numerator of written durations if necessary
        if (forbidden_duration is not None and
            forbidden_duration <= duration):
            denominators = [
                2 * forbidden_duration.denominator,
                duration.denominator,
                ]
            denominator = abjad.mathtools.least_common_multiple(*denominators)
            forbidden_duration = abjad.NonreducedFraction(forbidden_duration)
            forbidden_duration = forbidden_duration.with_denominator(
                denominator)
            duration = abjad.NonreducedFraction(duration)
            duration = duration.with_denominator(denominator)
            forbidden_numerator = forbidden_duration.numerator
            assert forbidden_numerator % 2 == 0
            preferred_numerator = forbidden_numerator / 2
        # make written duration numerators
        numerators = []
        parts = abjad.mathtools.partition_integer_into_canonic_parts(
            duration.numerator)
        if (forbidden_duration is not None and
            forbidden_duration <= duration):
            for part in parts:
                if forbidden_numerator <= part:
                    better_parts = LeafMaker._partition_less_than_double(
                        part,
                        preferred_numerator,
                        )
                    numerators.extend(better_parts)
                else:
                    numerators.append(part)
        else:
            numerators = parts
        # reverse numerators if necessary
        if not decrease_monotonic:
            numerators = list(reversed(numerators))
        # make one leaf per written duration
        result = []
        for numerator in numerators:
            written_duration = abjad.Duration(
                numerator,
                duration.denominator,
                )
            if pitches is not None:
                arguments = (pitches, written_duration)
            else:
                arguments = (written_duration, )
            result.append(class_(*arguments))
        result = abjad.select(result)
        # apply tie spanner if required
        if tie_parts and 1 < len(result):
            if not issubclass(class_, (abjad.Rest, abjad.Skip)):
                tie = abjad.Tie(repeat=repeat_ties)
                abjad.attach(tie, result)
        # return result
        return result

    @staticmethod
    def _partition_less_than_double(n, m):
        import abjad
        assert abjad.mathtools.is_positive_integer_equivalent_number(n)
        assert abjad.mathtools.is_positive_integer_equivalent_number(m)
        n, m = int(n), int(m)
        result = []
        current_value = n
        double_m = 2 * m
        while double_m <= current_value:
            result.append(m)
            current_value -= m
        result.append(current_value)
        return tuple(result)

    ### PUBLIC PROPERTIES ###

    @property
    def decrease_monotonic(self):
        """
        Is true when durations decrease monotonically.

        Returns true, false or none.
        """
        return self._decrease_monotonic

    @property
    def forbidden_duration(self):
        """
        Gets forbidden written duration.

        Returns duration or none.
        """
        return self._forbidden_duration

    @property
    def metrical_hierarchy(self):
        """
        Gets metrical hierarchy.

        Returns metrical hierarchy or none.
        """
        return self._metrical_hierarchy

    @property
    def skips_instead_of_rests(self):
        """
        Is true when skips appear in place of rests.

        Returns true, false or none.
        """
        return self._skips_instead_of_rests

    @property
    def repeat_ties(self) -> bool:
        """
        Is true when ties are repeat ties.
        """
        return self._repeat_ties

    @property
    def use_multimeasure_rests(self):
        """
        Is true when rests are multimeasure.

        Returns true, false or none.
        """
        return self._use_multimeasure_rests



import collections
from abjad import markups
from abjad.system.AbjadValueObject import AbjadValueObject


class Transposition(AbjadValueObject):
    """
    Transposition operator.

    ..  container:: example

        >>> abjad.Transposition()
        Transposition(n=0)

    ..  container:: example

        >>> abjad.Transposition(n=2)
        Transposition(n=2)

    Object model of twelve-tone transposition operator.
    """

    ### CLASS VARIABLES ##

    __slots__ = (
        '_n',
        )

    ### INITIALIZER ###

    def __init__(self, *, n=0):
        self._n = n

    ### SPECIAL METHODS ###

    def __add__(self, operator):
        r"""
        Composes transposition and `operator`.

        ..  container:: example

            Example segment:

            >>> items = [0, 2, 4, 5]
            >>> segment = abjad.PitchClassSegment(items=items)
            >>> abjad.show(segment) # doctest: +SKIP

            Example operators:

            >>> T_1 = abjad.Transposition(n=1)
            >>> T_3 = abjad.Transposition(n=3)

        ..  container:: example

            Successive transposition:

            >>> operator = T_1 + T_3
            >>> str(operator)
            'T1T3'

            >>> segment_ = operator(segment)
            >>> abjad.show(segment_) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment_.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    e'8
                    fs'8
                    af'8
                    a'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        ..  container:: example

            Same as above because transposition commutes:

            >>> operator = T_3 + T_1
            >>> str(operator)
            'T3T1'

            >>> segment_ = operator(segment)
            >>> abjad.show(segment_) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment_.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    e'8
                    fs'8
                    af'8
                    a'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        Returns compound operator.
        """
        import abjad
        return abjad.CompoundOperator._compose_operators(self, operator)

    def __call__(self, argument):
        """
        Calls transposition on `argument`.

        ..  container:: example

            Transposes pitch-class:

            >>> transposition = abjad.Transposition(n=2)
            >>> pitch_class = abjad.NumberedPitchClass(1)
            >>> transposition(pitch_class)
            NumberedPitchClass(3)

        ..  container:: example

            Transposes pitch:

            >>> transposition = abjad.Transposition(n=2)
            >>> pitch = abjad.NumberedPitch(15)
            >>> transposition(pitch)
            NumberedPitch(17)

        ..  container:: example

            Transposes list of pitches:

            >>> transposition = abjad.Transposition(n=2)
            >>> pitches = [abjad.NumberedPitch(_) for _ in [15, 16]]
            >>> transposition(pitches)
            [NumberedPitch(17), NumberedPitch(18)]

        Returns new object with type equal to that of `argument`.
        """
        if hasattr(argument, 'transpose'):
            result = argument.transpose(self.n)
        elif isinstance(argument, collections.Iterable):
            items = []
            for item in argument:
                item = item.transpose(self.n)
                items.append(item)
            result = type(argument)(items)
        else:
            message = 'do not know how to transpose: {!r}.'
            message = message.format(argument)
            raise TypeError(message)
        return result

    def __radd__(self, operator):
        """
        Right-addition not defined on transposition.

        ..  container:: example

            >>> abjad.Transposition().__radd__(abjad.Transposition())
            Traceback (most recent call last):
                ...
            NotImplementedError: right-addition not defined on Transposition.

        Raises not implemented error.
        """
        message = 'right-addition not defined on {}.'
        message = message.format(type(self).__name__)
        raise NotImplementedError(message)

    def __str__(self):
        """
        Gets string representation of operator.

        ..  container:: example

            >>> str(abjad.Transposition())
            'T0'

        ..  container:: example

            >>> str(abjad.Transposition(n=2))
            'T2'

        """
        string = 'T{}'
        string = string.format(self.n)
        return string

    ### PRIVATE METHODS ###

    def _get_markup(self, direction=None):
        operator = markups.Markup('T', direction=None)
        subscript = markups.Markup(self.n).sub()
        markup = markups.Markup.concat([operator, subscript])
        return markup

    def _is_identity_operator(self):
        if self.n == 0:
            return True
        return False

    ### PUBLIC PROPERTIES ###

    @property
    def n(self):
        """
        Gets index of transposition.

        ..  container:: example

            >>> transposition = abjad.Transposition()
            >>> transposition.n
            0

        ..  container:: example

            >>> transposition = abjad.Transposition(n=2)
            >>> transposition.n
            2

        Set to integer, interval or none.
        """
        return self._n



import copy
from abjad.pitch.NamedPitch import NamedPitch
from abjad.utilities.Duration import Duration
from .DrumNoteHead import DrumNoteHead
from .Leaf import Leaf
from .NoteHead import NoteHead


class Note(Leaf):
    """
    Note.

    ..  container:: example

        >>> note = abjad.Note("cs''8.")
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            cs''8.


    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Leaves'

    __slots__ = (
        '_note_head',
        )

    ### INITIALIZER ###

    def __init__(self, *arguments):
        import abjad
        from abjad.ly import drums
        assert len(arguments) in (0, 1, 2)
        if len(arguments) == 1 and isinstance(arguments[0], str):
            string = '{{ {} }}'.format(arguments[0])
            parsed = abjad.parse(string)
            assert len(parsed) == 1 and isinstance(parsed[0], Leaf)
            arguments = [parsed[0]]
        is_cautionary = False
        is_forced = False
        is_parenthesized = False
        if len(arguments) == 1 and isinstance(arguments[0], Leaf):
            leaf = arguments[0]
            written_pitch = None
            written_duration = leaf.written_duration
            if 'written_pitch' in dir(leaf):
                written_pitch = leaf.note_head.written_pitch
                is_cautionary = leaf.note_head.is_cautionary
                is_forced = leaf.note_head.is_forced
                is_parenthesized = leaf.note_head.is_parenthesized
            elif 'written_pitches' in dir(leaf):
                written_pitches = [x.written_pitch for x in leaf.note_heads]
                if written_pitches:
                    written_pitch = written_pitches[0]
                    is_cautionary = leaf.note_heads[0].is_cautionary
                    is_forced = leaf.note_heads[0].is_forced
                    is_parenthesized = leaf.note_heads[0].is_parenthesized
        elif len(arguments) == 2:
            written_pitch, written_duration = arguments
        elif len(arguments) == 0:
            written_pitch = 'C4'
            written_duration = Duration(1, 4)
        else:
            message = 'can not initialize note from {!r}.'
            raise ValueError(message.format(arguments))
        Leaf.__init__(self, written_duration)
        if written_pitch is not None:
            if written_pitch not in drums:
                self.note_head = NoteHead(
                    written_pitch=written_pitch,
                    is_cautionary=is_cautionary,
                    is_forced=is_forced,
                    is_parenthesized=is_parenthesized,
                    )
            else:
                self.note_head = DrumNoteHead(
                    written_pitch=written_pitch,
                    is_cautionary=is_cautionary,
                    is_forced=is_forced,
                    is_parenthesized=is_parenthesized,
                    )
        else:
            self.note_head = None
        if len(arguments) == 1 and isinstance(arguments[0], Leaf):
            self._copy_override_and_set_from_leaf(arguments[0])

    ### SPECIAL METHODS ###

    def __getnewargs__(self):
        """
        Gets new arguments.

        Returns tuple.
        """
        return (self.written_pitch, self.written_duration)

    ### PRIVATE METHODS ###

    def _get_body(self):
        formatted_duration = self._get_formatted_duration()
        if self.note_head is not None:
            string = self.note_head._get_lilypond_format(
                formatted_duration=formatted_duration
                )
        else:
            string = formatted_duration
        return [string]

    def _get_compact_representation(self):
        return self._get_body()[0]

    def _get_compact_representation_with_tie(self):
        logical_tie = self._get_logical_tie()
        if 1 < len(logical_tie) and self is not logical_tie[-1]:
            return '{} ~'.format(self._get_body()[0])
        else:
            return self._get_body()[0]

    def _get_sounding_pitch(self):
        import abjad
        if 'sounding pitch' in abjad.inspect(self).indicators(str):
            return self.written_pitch
        else:
            instrument = self._get_effective(abjad.Instrument)
            if instrument:
                sounding_pitch = instrument.middle_c_sounding_pitch
            else:
                sounding_pitch = NamedPitch('C4')
            interval = NamedPitch('C4') - sounding_pitch
            sounding_pitch = interval.transpose(self.written_pitch)
            return sounding_pitch

    ### PUBLIC PROPERTIES ###

    @property
    def note_head(self) -> NoteHead:
        """
        Gets and sets note-head.

        .. container:: example

            >>> note = abjad.Note("cs''8.")
            >>> note.note_head
            NoteHead("cs''")

            >>> abjad.show(note) # doctest: +SKIP
            
            ..  docs::
            
                >>> abjad.f(note)
                cs''8.
                
            >>> note.note_head = 'D5'
            >>> note.note_head
            NoteHead("d''")

            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                d''8.

        """
        return self._note_head

    @note_head.setter
    def note_head(self, argument):
        if isinstance(argument, type(None)):
            self._note_head = None
        elif isinstance(argument, NoteHead):
            self._note_head = argument
        else:
            note_head = NoteHead(client=self, written_pitch=argument)
            self._note_head = note_head

    @property
    def written_duration(self) -> Duration:
        """
        Gets and sets written duration.

        ..  container:: example

            >>> note = abjad.Note("cs''8.")
            >>> note.written_duration
            Duration(3, 16)

            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''8.

            >>> note.written_duration = (1, 16)
            >>> note.written_duration
            Duration(1, 16)

            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''16

        Returns duration
        """
        return Leaf.written_duration.fget(self)

    @written_duration.setter
    def written_duration(self, argument):
        return Leaf.written_duration.fset(self, argument)

    @property
    def written_pitch(self) -> NamedPitch:
        """
        Gets and sets written pitch.

        ..  container:: example

            >>> note = abjad.Note("cs''8.")
            >>> note.written_pitch
            NamedPitch("cs''")

            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''8.

            >>> note.written_pitch = 'D5'
            >>> note.written_pitch
            NamedPitch("d''")

            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                d''8.

        Returns named pitch.
        """
        if self.note_head is not None:
            return self.note_head.written_pitch
        else:
            return None

    @written_pitch.setter
    def written_pitch(self, argument):
        if argument is None:
            if self.note_head is not None:
                self.note_head.written_pitch = None
        else:
            if self.note_head is None:
                self.note_head = NoteHead(self, written_pitch=None)
            else:
                pitch = NamedPitch(argument)
                self.note_head.written_pitch = pitch

    ### PUBLIC METHODS ###

    @staticmethod
    def from_pitch_and_duration(pitch, duration):
        """
        Makes note from ``pitch`` and ``duration``.

        ..  container:: example

            >>> note = abjad.Note.from_pitch_and_duration('C#5', (3, 16))
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''8.

        """
        note = Note(pitch, duration)
        return note



from abjad.system.AbjadObject import AbjadObject


class SyntaxNode(AbjadObject):
    """
    A node in an abstract syntax tree (AST).

    Not composer-safe.

    Used internally by LilyPondParser.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        'type',
        'value',
        )

    ### INTIAILIZER ###

    def __init__(self, type=None, value=None):
        self.type = type
        self.value = value

    ### SPECIAL METHODS ###

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        Returns item or slice.
        """
        if isinstance(self.value, (list, tuple)):
            return self.value.__getitem__(argument)
        message = 'can not get: {!r}.'
        message = message.format(argument)
        raise Exception(message)

    def __len__(self):
        """
        Length of syntax node.

        Returns nonnegative integer.
        """
        if isinstance(self.value, (list, tuple)):
            return len(self.value)
        message = 'value must be list or tuple.'
        raise Exception(message)

    def __repr__(self):
        """
        Gets interpreter representation of syntax node.

        Returns string.
        """
        return '{}({}, {})'.format(
            type(self).__name__,
            self.type,
            type(self.value),
            )

    def __str__(self):
        """
        String representation of syntax node.

        Returns string.
        """
        return '\n'.join(self._format(self))

    ### PRIVATE METHODS ###

    def _format(self, obj, indent=0):
        space = '.  ' * indent
        result = []

        if isinstance(obj, type(self)):
            if isinstance(obj.value, (list, tuple)):
                result.append('%s<%s>: [' % (space, obj.type))
                for x in obj.value:
                    result.extend(self._format(x, indent + 1))
                result[-1] += ' ]'
            else:
                result.append('%s<%s>: %r' % (space, obj.type, obj.value))

        elif isinstance(obj, (list, tuple)):
            result.append('%s[' % space)
            for x in obj:
                result.extend(self._format(x, indent + 1))
            result[-1] += ' ]'

        else:
            result.append('%s%r' % (space, obj))

        return result



import typing
from abjad import enums
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class Arpeggio(AbjadValueObject):
    r"""
    Arpeggio.

    ..  container:: example

        Without direction arrow:

        >>> chord = abjad.Chord("<c' e' g' c''>4")
        >>> arpeggio = abjad.Arpeggio()
        >>> abjad.attach(arpeggio, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(chord)
            <c' e' g' c''>4
            \arpeggio

    ..  container:: example

        With direction arrow:

        >>> chord = abjad.Chord("<c' e' g' c''>4")
        >>> arpeggio = abjad.Arpeggio(direction=abjad.Down)
        >>> abjad.attach(arpeggio, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(chord)
            \arpeggioArrowDown
            <c' e' g' c''>4
            \arpeggio

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_direction',
        '_tweaks',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        direction: enums.VerticalAlignment = None,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        if direction is not None:
            assert direction in (enums.Up, enums.Down, enums.Center)
        self._direction = direction
        self._tweaks = None
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return r'\arpeggio'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.articulations.extend(tweaks)
        bundle.after.articulations.append(r'\arpeggio')
        if self.direction in (enums.Up, enums.Down):
            if self.direction is enums.Up:
                command = r'\arpeggioArrowUp'
            else:
                command = r'\arpeggioArrowDown'
            bundle.before.commands.append(command)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def direction(self) -> typing.Optional[enums.VerticalAlignment]:
        """
        Gets direction of arpeggio.

        ..  container:: example

            Without direction arrow:

            >>> arpeggio = abjad.Arpeggio()
            >>> arpeggio.direction is None
            True

        ..  container:: example

            With direction arrow:

            >>> arpeggio = abjad.Arpeggio(direction=abjad.Down)
            >>> arpeggio.direction
            Down

        """
        return self._direction

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> chord = abjad.Chord("<c' e' g' c''>4")
            >>> arpeggio = abjad.Arpeggio()
            >>> abjad.tweak(arpeggio).color = 'blue'
            >>> abjad.attach(arpeggio, chord)
            >>> abjad.show(chord) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(chord)
                <c' e' g' c''>4
                - \tweak color #blue
                \arpeggio

        """
        return self._tweaks



import collections
import itertools
import ply  # type: ignore
from abjad import indicators as abjad_indicators
from abjad import lilypondfile as abjad_lilypondfile
from abjad import markups as abjad_markups
from abjad import pitch as abjad_pitch
from abjad import core
from abjad import spanners as abjad_spanners
from abjad.system import Parser
from ._parse import _parse
from ._parse_debug import _parse_debug
from abjad.top.attach import attach
from abjad.top.detach import detach
from abjad.top.select import select


# apply monkey patch
ply.yacc.LRParser._lilypond_patch_parse = _parse
ply.yacc.LRParser._lilypond_patch_parse_debug = _parse_debug


class LilyPondParser(Parser):
    r"""A LilyPond syntax parser.

    ..  container:: example

        >>> parser = abjad.parser.LilyPondParser()
        >>> string = r"\new Staff { c'4 ( d'8 e' fs'2) \fermata }"
        >>> staff = parser(string)
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            (
            d'8
            e'8
            fs'2
            -\fermata
            )
        }

    ..  container:: example

        The LilyPond parser understands most spanners, articulations and
        dynamics:

        >>> string = r'''
        ... \new Staff {
        ...     c'8 \f \> (
        ...     d' -_ [
        ...     e' ^>
        ...     f' \ppp \<
        ...     g' \startTrillSpan \(
        ...     a' \)
        ...     b' ] \stopTrillSpan
        ...     c'' ) \accent \sfz
        ... }
        ... '''
        >>> staff = parser(string)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'8
                \f
                \>
                (
                d'8
                -\portato
                [
                e'8
                ^\accent
                f'8
                \ppp
                \<
                g'8
                \(
                \startTrillSpan
                a'8
                \)
                b'8
                ]
                \stopTrillSpan
                c''8
                \sfz
                -\accent
                )
            }

    ..  container:: example

        The LilyPond parser understands contexts and markup:

        >>> string = r'''\new Score <<
        ...     \new Staff = "Treble Staff" {
        ...         \new Voice = "Treble Voice" {
        ...             c' ^\markup { \bold Treble! }
        ...         }
        ...     }
        ...     \new Staff = "Bass Staff" {
        ...         \new Voice = "Bass Voice" {
        ...             \clef bass
        ...             c, _\markup { \italic Bass! }
        ...         }
        ...     }
        ... >>
        ... '''
        >>> score = parser(string)
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \context Staff = "Treble Staff"
                {
                    \context Voice = "Treble Voice"
                    {
                        c'4
                        ^ \markup {
                            \bold
                                Treble!
                            }
                    }
                }
                \context Staff = "Bass Staff"
                {
                    \context Voice = "Bass Voice"
                    {
                        \clef "bass"
                        c,4
                        _ \markup {
                            \italic
                                Bass!
                            }
                    }
                }
            >>

    ..  container:: example

        The LilyPond parser also understands certain aspects of LilyPond file
        layouts, such as header blocks:

        >>> string = r'''
        ... \header {
        ...     composername = "Foo von Bar"
        ...     composer = \markup { by \bold \composername }
        ...     title = \markup { The ballad of \composername }
        ...     tagline = \markup { "" }
        ... }
        ... \score {
        ...     \new Staff {
        ...         \time 3/4
        ...         g' ( b' d'' )
        ...         e''4. ( c''8 c'4 )
        ...     }
        ... }
        ... '''
        >>> blocks = parser(string)
        >>> abjad.show(blocks) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(blocks) # doctest: +SKIP
            % 2017-07-11 15:13
            <BLANKLINE>
            \version "2.19.63"
            \language "english"
            <BLANKLINE>
            \header {
                composername = #"Foo von Bar"
                composer = \markup {
                    by
                    \bold
                        "Foo von Bar"
                    }
                title = \markup {
                    The
                    ballad
                    of
                    "Foo von Bar"
                    }
                tagline = \markup {}
            }
            <BLANKLINE>
            \score {
                \new Staff {
                    \time 3/4
                    g'4 (
                    b'4
                    d''4 )
                    e''4. (
                    c''8
                    c'4 )
                }
            }

    ..  container:: example

        The LilyPond parser supports a small number of LilyPond music
        functions, such as \relative and \transpose.

        ..  note::

            Music functions which mutate the score during compilation result in a
            normalized Abjad score structure. The resulting structure corresponds
            to the music as it appears on the page, rather than as it was input to
            the parser:

        >>> string = r'''
        ... \new Staff \relative c {
        ...     c32 d e f g a b c d e f g a b c d e f g a b c
        ... }
        ... '''
        >>> staff = parser(string)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c32
                d32
                e32
                f32
                g32
                a32
                b32
                c'32
                d'32
                e'32
                f'32
                g'32
                a'32
                b'32
                c''32
                d''32
                e''32
                f''32
                g''32
                a''32
                b''32
                c'''32
            }

    ..  container:: example

        The LilyPond parser defaults to English note names, but any of the
        other languages supported by LilyPond may be used:

        >>> parser = abjad.parser.LilyPondParser('nederlands')
        >>> string = '{ c des e fis }'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c4
                df4
                e4
                fs4
            }

    Briefly, LilyPondParser understands theses aspects of LilyPond syntax:

    - Notes, chords, rests, skips and multi-measure rests
    - Durations, dots, and multipliers
    - All pitchnames, and octave ticks
    - Simple markup (i.e. ``c'4 ^ "hello!"``)
    - Most articulations
    - Most spanners, incl. beams, slurs, phrasing slurs, ties, and glissandi
    - Most context types via ``\new`` and ``\context``,
      as well as context ids (i.e. ``\new Staff = "foo" { }``)
    - Variable assigns (ie ``global = { \time 3/4 } \new Staff { \global }``)
    - Many music functions:
        - ``\acciaccatura``
        - ``\appoggiatura``
        - ``\bar``
        - ``\breathe``
        - ``\clef``
        - ``\grace``
        - ``\key``
        - ``\transpose``
        - ``\language``
        - ``\makeClusters``
        - ``\mark``
        - ``\oneVoice``
        - ``\relative``
        - ``\skip``
        - ``\slashedGrace``
        - ``\time``
        - ``\times``
        - ``\transpose``
        - ``\voiceOne``, ``\voiceTwo``, ``\voiceThree``, ``\voiceFour``

    LilyPondParser currently **DOES NOT** understand many other aspects
    of LilyPond syntax:

    - ``\markup``
    - ``\book``, ``\bookpart``, ``\header``, ``\layout``, ``\midi``, ``\paper``
    - ``\repeat`` and ``\alternative``
    - Lyrics
    - ``\chordmode``, ``\drummode`` or ``\figuremode``
    - Property operations, such as ``\override``,
      ``\revert``, ``\set``, ``\unset``, and ``\once``
    - Music functions which generate or extensively mutate musical structures
    - Embedded Scheme statements (anything beginning with ``#``)
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_chord_pitch_orders',
        '_current_module',
        '_default_duration',
        '_default_language',
        '_guile',
        '_language_pitch_names',
        '_last_chord',
        '_lexdef',
        '_markup_functions',
        '_markup_list_functions',
        '_pitch_names',
        '_repeated_chords',
        '_scope_stack',
        '_syndef',
        )

    ### INITIALIZER ###

    def __init__(self, default_language='english', debug=False):
        from abjad import parser as abjad_parser
        from abjad.ly.current_module import current_module
        from abjad.ly.language_pitch_names import language_pitch_names
        from abjad.ly.markup_functions import markup_functions
        from abjad.ly.markup_functions import markup_list_functions

        # LilyPond emulation data
        self._guile = abjad_parser.GuileProxy(self)
        self._current_module = current_module
        self._language_pitch_names = language_pitch_names
        self._markup_functions = markup_functions
        self._markup_list_functions = markup_list_functions
        self.default_language = default_language

        # attach parser and lexer rules
        self._lexdef = abjad_parser.LilyPondLexicalDefinition(self)
        self._syndef = abjad_parser.LilyPondSyntacticalDefinition(self)

        # build PLY parser and lexer
        Parser.__init__(self, debug=debug)

        self._reset_parser_variables()

    ### SPECIAL METHODS ###

    def __call__(self, input_string):
        """
        Calls LilyPond parser on `input_string`.

        Returns Abjad components.
        """
        self._reset_parser_variables()
        if self._debug:
            result = self._parser._lilypond_patch_parse_debug(
                input_string,
                lexer=self._lexer,
                debug=self._logger,
                )
        else:
            result = self._parser._lilypond_patch_parse(
                input_string,
                lexer=self._lexer,
                )
        if isinstance(result, core.Container):
            self._apply_spanners(result)
        elif isinstance(result, abjad_lilypondfile.LilyPondFile):
            for x in result.items:
                if isinstance(x, core.Container):
                    self._apply_spanners(x)
                elif isinstance(x, abjad_lilypondfile.Block) and \
                    x.name == 'score':
                    for y in x.items:
                        self._apply_spanners(y)
        return result

    ### PRIVATE METHODS ###

    def _apply_spanners(self, music):
        import abjad
        # get local reference to methods
        _get_span_events = self._get_span_events
        _span_event_name_to_spanner_class = \
            self._span_event_name_to_spanner_class

        # dictionary of spanner classes and instances
        all_spanners = {}

        # traverse all leaves
        leaves = select(music).leaves()
        first_leaf = None
        if leaves:
            first_leaf = leaves[0]
        pairs = abjad.sequence(leaves).nwise(wrapped=True)
        for leaf, next_leaf in pairs:

            span_events = _get_span_events(leaf)
            directed_events = {}

            # sort span events into directed and undirected groups
            for span_event in span_events:
                spanner_class = _span_event_name_to_spanner_class(
                    span_event.name)

                # group directed span events by their Abjad spanner class
                if hasattr(span_event, 'span_direction'):
                    if spanner_class not in directed_events:
                        directed_events[spanner_class] = [span_event]
                    else:
                        directed_events[spanner_class].append(span_event)
                    if spanner_class not in all_spanners:
                        all_spanners[spanner_class] = []

                # or apply undirected event immediately (i.e. ties, glisses)
                # so long as we are not wrapping yet
                elif next_leaf is not first_leaf:
                    previous_spanners = [
                        x for x in leaf._get_spanners()
                        if isinstance(x, spanner_class)
                        ]
                    if previous_spanners:
                        previous_spanners[0]._append(next_leaf)
                    else:
                        if (hasattr(span_event, 'direction') and
                            hasattr(spanner_class, 'direction')):
                            direction = span_event.direction
                            spanner = spanner_class(direction=direction)
                            selection = abjad.select([leaf, next_leaf]).leaves()
                            attach(spanner, selection)
                        else:
                            spanner = spanner_class()
                            selection = abjad.select([leaf, next_leaf]).leaves()
                            attach(spanner, selection)

                # otherwise throw an error
                else:
                    message = 'unterminated {} at {}.'
                    message = message.format(spanner_class.__name__, leaf)
                    raise Exception(message)

            # check for dynamics and terminate any hairpin
            dynamics = leaf._get_indicators(abjad_indicators.Dynamic)
            if dynamics and abjad_spanners.Hairpin in all_spanners and \
                all_spanners[abjad_spanners.Hairpin]:
                all_spanners[abjad_spanners.Hairpin][0]._append(leaf)
                all_spanners[abjad_spanners.Hairpin].pop()

            # loop through directed events, handling each as necessary
            for spanner_class, events in directed_events.items():

                starting_events, stopping_events = [], []
                for x in events:
                    if x.span_direction == 'start':
                        starting_events.append(x)
                    else:
                        stopping_events.append(x)

                if spanner_class is abjad_spanners.Beam:
                    # A beam may begin and end on the same leaf
                    # but only one beam spanner may cover any given leaf,
                    # and starting events are processed before ending ones
                    for event in starting_events:
                        if all_spanners[spanner_class]:
                            raise Exception('already have beam.')
                        if hasattr(event, 'direction'):
                            spanner_ = spanner_class(
                                beam_lone_notes=True,
                                direction=event.direction,
                                )
                            all_spanners[spanner_class].append(spanner_)
                        else:
                            spanner_ = spanner_class(
                                beam_lone_notes=True,
                                )
                            all_spanners[spanner_class].append(spanner_)
                    for _ in stopping_events:
                        if all_spanners[spanner_class]:
                            all_spanners[spanner_class][0]._append(leaf)
                            all_spanners[spanner_class].pop()

                elif spanner_class is abjad_spanners.Hairpin:
                    # Dynamic events can be ended many times,
                    # but only one may start on a given leaf,
                    # and the event must start and end on separate leaves.
                    # If a hairpin already exists and another starts,
                    # the preexistent spanner is ended.
                    for _ in stopping_events:
                        if all_spanners[spanner_class]:
                            all_spanners[spanner_class][0]._append(leaf)
                            all_spanners[spanner_class].pop()
                    if 1 == len(starting_events):
                        if all_spanners[spanner_class]:
                            all_spanners[spanner_class][0]._append(leaf)
                            all_spanners[spanner_class].pop()
                        shape = '<'
                        event = starting_events[0]
                        if event.name == 'DecrescendoEvent':
                            shape = '>'
                        if hasattr(event, 'direction'):
                            spanner = spanner_class(
                                descriptor=shape,
                                direction=event.direction,
                                )
                            all_spanners[spanner_class].append(spanner)
                        else:
                            spanner = spanner_class(descriptor=shape)
                            all_spanners[spanner_class].append(spanner)
                    elif 1 < len(starting_events):
                        raise Exception('simultaneous dynamic span events.')

                elif spanner_class in [
                    abjad_spanners.Slur,
                    abjad_spanners.PhrasingSlur,
                    abjad_spanners.TextSpanner,
                    abjad_spanners.TrillSpanner,
                    ]:
                    # these engravers process stop events before start events,
                    # they must contain more than one leaf, however,
                    # yhey can stop on a leaf and start on the same leaf.
                    for _ in stopping_events:
                        if all_spanners[spanner_class]:
                            all_spanners[spanner_class][0]._append(leaf)
                            all_spanners[spanner_class].pop()
                        else:
                            message = 'can not end {}.'
                            message = message.format(spanner_class.__name__)
                            raise Exception(message)
                    for event in starting_events:
                        if not all_spanners[spanner_class]:
                            if hasattr(event, 'direction') and \
                                hasattr(spanner_class, 'direction'):
                                all_spanners[spanner_class].append(
                                    spanner_class(direction=event.direction))
                            else:
                                all_spanners[spanner_class].append(
                                    spanner_class())
                        else:
                            message = 'already have {}.'
                            message = message.format(spanner_class.__name__)
                            raise Exception(message)

                elif spanner_class is abjad_spanners.HorizontalBracket:
                    # Brackets can nest, meaning
                    # multiple brackets can begin or end on a leaf
                    # but can not both begin and end on the same leaf
                    # and therefore a bracket can not cover a single leaf
                    has_starting_events = bool(len(starting_events))
                    for _ in starting_events:
                        all_spanners[spanner_class].append(spanner_class())
                    if stopping_events:
                        if not has_starting_events:
                            for _ in stopping_events:
                                if all_spanners[spanner_class]:
                                    all_spanners[spanner_class][-1]._append(
                                        leaf)
                                    all_spanners[spanner_class].pop()
                                else:
                                    message = 'do not have that many brackets.'
                                    raise Exception(message)
                        else:
                            message = 'conflicting note group events.'
                            raise Exception(message)

            # append leaf to all tracked spanners,
            for spanner_class, instances in all_spanners.items():
                for instance in instances:
                    instance._append(leaf)

        # check for unterminated spanners
        for spanner_class, instances in all_spanners.items():
            if instances:
                message = 'unterminated {}.'
                message = message.format(spanner_class.__name__)
                raise Exception(message)

    def _assign_variable(self, identifier, value):
        self._scope_stack[-1][identifier] = value

    def _backup_token(self, token_type, token_value):
        if self._debug:
            self._logger.info('Extra  : Backing up')

        # push the current lookahead back onto the lookaheadstack
        self._push_extra_token(self._parser.lookahead)

        # create the backup token, set as new lookahead
        backup = ply.lex.LexToken()
        backup.type = 'BACKUP'
        backup.value = '(backed-up?)'
        backup.lexpos = 0
        backup.lineno = 0
        self._parser.lookahead = backup

        if token_type:
            token = ply.lex.LexToken()
            token.type = token_type
            token.value = token_value
            token.lexpos = 0
            token.lineno = 0
            self._push_extra_token(token)

    def _construct_context_specced_music(
        self,
        context,
        optional_id,
        optional_context_mod,
        music,
        ):
        known_contexts = {
            'ChoirStaff': core.StaffGroup,
            'GrandStaff': core.StaffGroup,
            'PianoStaff': core.StaffGroup,
            'Score': core.Score,
            'Staff': core.Staff,
            'StaffGroup': core.StaffGroup,
            'Voice': core.Voice,
            }
        lilypond_type = context
        if context in known_contexts:
            context = known_contexts[context]([])
        else:
            message = 'context type {!r} not supported.'
            message = message.format(context)
            raise Exception(message)
        if lilypond_type in ('GrandStaff', 'PianoStaff'):
            context.lilypond_type = lilypond_type
        if optional_id is not None:
            context.name = optional_id
        if optional_context_mod is not None:
            for x in optional_context_mod:
                print(x)
            # TODO: impelement context mods on contexts
            pass
        context.is_simultaneous = music.is_simultaneous
        # add children
        while len(music):
            component = music.pop(0)
            context.append(component)
        for wrapper in music._wrappers:
            attach(wrapper, context)
        return context

    def _construct_sequential_music(self, music):
        # indicator sorting could be rewritten into a single list using tuplets
        # with t[0] being 'forward' or 'backward' and t[1] being the indicator
        # as this better preserves attachment order. Not clear if we need it.
        container = core.Container()
        previous_leaf = None
        apply_forward = []
        apply_backward = []
        # sort events into forward or backwards attaching
        # and attach them to the proper leaf
        for x in music:
            if isinstance(x, core.Component) \
                and not isinstance(x, core.GraceContainer):
                for indicator in apply_forward:
                    attach(indicator, x)
                if previous_leaf:
                    for indicator in apply_backward:
                        attach(indicator, previous_leaf)
                else:
                    for indicator in apply_backward:
                        attach(indicator, x)
                apply_forward[:] = []
                apply_backward[:] = []
                previous_leaf = x
                container.append(x)
            else:
                if isinstance(x, (
                    abjad_indicators.BarLine,
                    )):
                    apply_backward.append(x)
                elif (
                    isinstance(x, abjad_indicators.LilyPondLiteral) and
                    x.name in (r'\break', r'\breathe', r'\pageBreak')
                    ):
                    apply_backward.append(x)
                else:
                    apply_forward.append(x)
        # attach remaining events to last leaf
        # or to the container itself if there were no leaves
        if previous_leaf:
            for indicator in apply_forward:
                attach(indicator, previous_leaf)
            for indicator in apply_backward:
                attach(indicator, previous_leaf)
        else:
            for indicator in apply_forward:
                attach(indicator, container)
            for indicator in apply_backward:
                attach(indicator, container)
        return container

    def _construct_simultaneous_music(self, music):
        from abjad import parser as abjad_parser
        def is_separator(x):
            if isinstance(x, abjad_parser.LilyPondEvent):
                if x.name == 'VoiceSeparator':
                    return True
            return False
        container = core.Container()
        container.is_simultaneous = True
        # check for voice separators
        groups = []
        for value, group in itertools.groupby(music, is_separator):
            if not value:
                groups.append(list(group))
        # without voice separators
        if 1 == len(groups):
            #assert all(isinstance(x, core.Context) for x in groups[0])
            container.extend(groups[0])
        # with voice separators
        else:
            for group in groups:
                container.append(
                    core.Voice(self._construct_sequential_music(group)[:]))
        return container

    @classmethod
    def _get_scheme_predicates(class_):
        from abjad import parser as abjad_parser
        return {
            'boolean?': lambda x: isinstance(x, bool),
            'cheap-list?': lambda x: isinstance(x, (list, tuple)),
            'cheap-markup?': lambda x: isinstance(x,
                abjad_markups.MarkupCommand),
            'fraction?': lambda x: isinstance(x, abjad_parser.LilyPondFraction),
            'integer?': lambda x: isinstance(x, int),
            'list?': lambda x: isinstance(x, (list, tuple)),
            'ly:duration?': lambda x: isinstance(x,
                abjad_parser.LilyPondDuration),
            'ly:music?': lambda x: isinstance(x, core.Component),
            'ly:pitch?': lambda x: isinstance(x, abjad_pitch.NamedPitch),
            'markup?': lambda x: isinstance(x, abjad_markups.MarkupCommand),
            'number-list?': lambda x: isinstance(x, (list, tuple)) and
                all(isinstance(y, (int, float)) for y in x),
            'number?': lambda x: isinstance(x, (int, float)),
            'real?': lambda x: isinstance(x, (int, float)),
            'string?': lambda x: isinstance(x, str),
            'void?': lambda x: isinstance(x, type(None)),
            # the following predicates have not yet been implemented in Abjad
            'hash-table?': lambda x: True,
            'list-or-symbol?': lambda x: True,
            'ly:dir?': lambda x: True,
            'ly:moment?': lambda x: True,
            'number-or-string?': lambda x: True,
            'number-pair?': lambda x: True,
            'optional?': lambda x: True,
            'pair?': lambda x: True,
            'procedure?': lambda x: True,
            'scheme?': lambda x: True,
            'string-or-pair?': lambda x: True,
            'symbol-or-boolean?': lambda x: True,
            'symbol?': lambda x: True,
            }

    def _get_span_events(self, leaf):
        import abjad
        annotation = abjad.inspect(leaf).annotation('spanners', [])
        abjad.detach(annotation, leaf)
        assert isinstance(annotation, list), repr(annotation)
        assert abjad.inspect(leaf).annotation('spanners') is None
        return annotation

    def _pop_variable_scope(self):
        if self._scope_stack:
            self._scope_stack.pop()

    def _process_post_events(self, leaf, post_events):
        import abjad
        for post_event in post_events:
            # TODO: the conditional logic here will have to change;
            #       post events like StemTremolo no longer implement _attach.
            #       Is there a way to identify spanner LilyPondEvents?
            #       Then we can just (blindly) attach all other post events.
            nonspanner_post_event_types = (
                abjad_indicators.Articulation,
                abjad_indicators.BarLine,
                abjad_indicators.Dynamic,
                abjad_indicators.LilyPondLiteral,
                abjad_indicators.StemTremolo,
                abjad_markups.Markup,
                )
            if hasattr(post_event, '_attach'):
                attach(post_event, leaf)
            elif isinstance(post_event, nonspanner_post_event_types):
                attach(post_event, leaf)
            else:
                annotation = abjad.inspect(leaf).annotation('spanners')
                if annotation is None:
                    annotation = []
                    abjad.annotate(leaf, 'spanners', annotation)
                annotation.append(post_event)

    def _push_extra_token(self, token):
        self._parser.lookaheadstack.append(token)

    def _push_variable_scope(self):
        self._scope_stack.append({})

    def _relex_lookahead(self):
        if not str(self._parser.lookahead) == '$end':
            difference = self._parser.lookahead.lexpos - self._lexer.lexpos
            self._lexer.skip(difference)
            self._parser.lookahead = None

    def _reparse_token(self, predicate, token_type, token_value):
        if self._debug:
            self._logger.info('Extra  : Reparsing')

        # push the current lookahead back onto the lookaheadstack
        self._push_extra_token(self._parser.lookahead)

        token = ply.lex.LexToken()
        token.type = token_type
        token.value = token_value
        token.lexpos = 0
        token.lineno = 0
        self._push_extra_token(token)

        reparse = ply.lex.LexToken()
        reparse.type = 'REPARSE'
        reparse.value = predicate
        reparse.lineno = 0
        reparse.lexpos = 0
        self._parser.lookahead = reparse

    def _reset_parser_variables(self):
        from abjad import parser as abjad_parser
        import abjad
        try:
            self._parser.restart()
        except:
            pass
        self._scope_stack = [{}]
        self._chord_pitch_orders = {}
        self._lexer.push_state('notes')
        self._default_duration = abjad_parser.LilyPondDuration(
            abjad.Duration(1, 4), None)
        self._last_chord = None
        # LilyPond's default!
        # self._last_chord = core.Chord(['c', 'g', "c'"], (1, 4))
        self._pitch_names = self._language_pitch_names[self.default_language]
        self._repeated_chords = {}

    def _resolve_event_identifier(self, identifier):
        from abjad import parser as abjad_parser
        # without any leading slash
        lookup = self._current_module[identifier]
        name = lookup['name']
        if name == 'ArticulationEvent':
            return abjad_indicators.Articulation(lookup['articulation-type'])
        elif name == 'AbsoluteDynamicEvent':
            return abjad_indicators.Dynamic(lookup['text'])
        elif name == 'LaissezVibrerEvent':
            return abjad_indicators.LilyPondLiteral(r'\laissezVibrer', 'after')
        elif name == 'LineBreakEvent':
            return abjad_indicators.LilyPondLiteral(r'\break')
        event = abjad_parser.LilyPondEvent(name)
        if 'span-direction' in lookup:
            if lookup['span-direction'] == -1:
                event.span_direction = 'start'
            else:
                event.span_direction = 'stop'
        return event

    def _resolve_identifier(self, identifier):
        for scope in reversed(self._scope_stack):
            if identifier in scope:
                return scope[identifier]
        return None

    def _span_event_name_to_spanner_class(self, name):
        spanners = {
            'BeamEvent': abjad_spanners.Beam,
            'CrescendoEvent': abjad_spanners.Hairpin,
            'DecrescendoEvent': abjad_spanners.Hairpin,
            'GlissandoEvent': abjad_spanners.Glissando,
            'NoteGroupingEvent': abjad_spanners.HorizontalBracket,
            'PhrasingSlurEvent': abjad_spanners.PhrasingSlur,
            'SlurEvent': abjad_spanners.Slur,
            'TextSpanEvent': abjad_spanners.TextSpanner,
            'TieEvent': abjad_spanners.Tie,
            'TrillSpanEvent': abjad_spanners.TrillSpanner,
            }
        if name in spanners:
            return spanners[name]
        message = 'can not associate a spanner class with {}.'
        message = message.format(name)
        raise Exception(message)

    def _test_scheme_predicate(self, predicate, value):
        predicates = self._get_scheme_predicates()
        if predicate in predicates:
            return predicates[predicate](value)
        return True

    @staticmethod
    def _transpose_enharmonically(pitch_a, pitch_b, pitch_c):
        r'''Transpose `pitch_c` by the distance between `pitch_b`
        and `pitch_a`.

        This function was reverse-engineered from LilyPond's source code.

        Returns named pitch.
        '''
        def normalize_alteration(step, alteration):
            while 2. < alteration:
                alteration -= step_size(step)
                step += 1.
            while alteration < -2.:
                step -= 1.
                alteration += step_size(step)
            return step, alteration

        def normalize_octave(octave, step):
            normalized_step = step % len(scale)
            octave += (step - normalized_step) / len(scale)
            return octave, normalized_step

        def step_size(step):
            normalized_step = step % len(scale)
            if normalized_step == 6:
                return 1.  # b to c
            return scale[normalized_step + 1] - scale[normalized_step]

        if not isinstance(pitch_a, abjad_pitch.NamedPitch):
            pitch_a = abjad_pitch.NamedPitch(pitch_a)
        if not isinstance(pitch_b, abjad_pitch.NamedPitch):
            pitch_b = abjad_pitch.NamedPitch(pitch_b)
        if not isinstance(pitch_c, abjad_pitch.NamedPitch):
            pitch_c = abjad_pitch.NamedPitch(pitch_c)
        scale = [0., 2., 4., 5., 7., 9., 11.]
        a_oct, a_step, a_alt = pitch_a.octave.number, \
            pitch_a._get_diatonic_pc_number(), pitch_a.accidental.semitones
        b_oct, b_step, b_alt = pitch_b.octave.number, \
            pitch_b._get_diatonic_pc_number(), pitch_b.accidental.semitones
        c_oct, c_step, c_alt = pitch_c.octave.number, \
            pitch_c._get_diatonic_pc_number(), pitch_c.accidental.semitones
        d_oct, d_step, d_alt, d_tones = b_oct - a_oct, b_step - a_step, \
            b_alt - a_alt, float(pitch_b.number) - float(pitch_a.number)
        tmp_alt = float(pitch_c.number) + d_tones
        # print 'TMP_ALT: %f' % tmp_alt
        new_oct = c_oct + d_oct
        new_step = c_step + d_step
        new_alt = c_alt
        # print 'NEW:', new_oct, new_step, new_alt
        new_step, new_alt = normalize_alteration(new_step, new_alt)
        new_oct, new_step = normalize_octave(new_oct, new_step)
        # print 'NEW(norm):', new_oct, new_step, new_alt
        octave_ticks = str(abjad_pitch.Octave(new_oct))
        pitch_class_name = \
            abjad_pitch.constants._diatonic_pc_number_to_diatonic_pc_name[
                new_step % 7]
        #pitch_class_name = str(abjad_pitch.NamedDiatonicPitchClass(
        #    int(new_step)))
        accidental = str(abjad_pitch.Accidental(new_alt))
        tmp_pitch = abjad_pitch.NamedPitch(
            pitch_class_name + accidental + octave_ticks)
        # print 'TMP(pitch): %r' % tmp_pitch
        new_alt += tmp_alt - float(tmp_pitch.number)
        # print 'NEW(alt): %f' % new_alt
        new_step, new_alt = normalize_alteration(new_step, new_alt)
        new_oct, new_step = normalize_octave(new_oct, new_step)
        # print 'NEW(norm):', new_oct, new_step, new_alt
        octave_ticks = str(abjad_pitch.Octave(new_oct))
        #pitch_class_name = str(abjad_pitch.NamedDiatonicPitchClass(
        #    int(new_step)))
        pitch_class_name = \
            abjad_pitch.constants._diatonic_pc_number_to_diatonic_pc_name[
                new_step % 7]
        accidental = str(abjad_pitch.Accidental(new_alt))
        return abjad_pitch.NamedPitch(
            pitch_class_name + accidental + octave_ticks)

    ### PUBLIC METHODS ###

    @staticmethod
    def list_known_contexts():
        r'''Lists all LilyPond contexts recognized by LilyPond parser.

        ..  container:: example

            >>> class_ = abjad.parser.LilyPondParser
            >>> for context in class_.list_known_contexts():
            ...     print(context)
            ...
            ChoirStaff
            ChordNames
            CueVoice
            Devnull
            DrumStaff
            DrumVoice
            Dynamics
            FiguredBass
            FretBoards
            Global
            GrandStaff
            GregorianTranscriptionStaff
            GregorianTranscriptionVoice
            KievanStaff
            KievanVoice
            Lyrics
            MensuralStaff
            MensuralVoice
            NoteNames
            NullVoice
            PetrucciStaff
            PetrucciVoice
            PianoStaff
            RhythmicStaff
            Score
            Staff
            StaffGroup
            TabStaff
            TabVoice
            VaticanaStaff
            VaticanaVoice
            Voice

        Returns list.
        '''
        from abjad.ly import contexts
        return sorted(contexts.keys())

    @staticmethod
    def list_known_dynamics():
        r'''Lists all dynamics recognized by LilyPond parser.

        ..  container:: example

            >>> class_ = abjad.parser.LilyPondParser
            >>> for dynamic in class_.list_known_dynamics():
            ...     print(dynamic)
            ...
            f
            ff
            fff
            ffff
            fffff
            fp
            fz
            mf
            mp
            p
            pp
            ppp
            pppp
            ppppp
            rfz
            sf
            sff
            sfp
            sfz
            sp
            spp

        Returns tuple.
        '''
        from abjad.ly import current_module
        result = []
        for key, value in current_module.items():
            if not isinstance(value, dict):
                continue
            if 'dynamic-event' in value.get('types', ()):
                result.append(key)
        result.sort()
        result = tuple(result)
        return result

    @staticmethod
    def list_known_grobs():
        r'''Lists all LilyPond grobs recognized by LilyPond parser.

        ..  container:: example

            >>> class_ = abjad.parser.LilyPondParser
            >>> for grob in class_.list_known_grobs():
            ...     print(grob)
            ...
            Accidental
            AccidentalCautionary
            AccidentalPlacement
            AccidentalSuggestion
            Ambitus
            AmbitusAccidental
            AmbitusLine
            AmbitusNoteHead
            Arpeggio
            BalloonTextItem
            BarLine
            BarNumber
            BassFigure
            BassFigureAlignment
            BassFigureAlignmentPositioning
            BassFigureBracket
            BassFigureContinuation
            BassFigureLine
            Beam
            BendAfter
            BreakAlignGroup
            BreakAlignment
            BreathingSign
            ChordName
            Clef
            ClefModifier
            ClusterSpanner
            ClusterSpannerBeacon
            CombineTextScript
            CueClef
            CueEndClef
            Custos
            DotColumn
            Dots
            DoublePercentRepeat
            DoublePercentRepeatCounter
            DoubleRepeatSlash
            DynamicLineSpanner
            DynamicText
            DynamicTextSpanner
            Episema
            Fingering
            FingeringColumn
            Flag
            FootnoteItem
            FootnoteSpanner
            FretBoard
            Glissando
            GraceSpacing
            GridLine
            GridPoint
            Hairpin
            HorizontalBracket
            InstrumentName
            InstrumentSwitch
            KeyCancellation
            KeySignature
            KievanLigature
            LaissezVibrerTie
            LaissezVibrerTieColumn
            LedgerLineSpanner
            LeftEdge
            LigatureBracket
            LyricExtender
            LyricHyphen
            LyricSpace
            LyricText
            MeasureCounter
            MeasureGrouping
            MelodyItem
            MensuralLigature
            MetronomeMark
            MultiMeasureRest
            MultiMeasureRestNumber
            MultiMeasureRestText
            NonMusicalPaperColumn
            NoteCollision
            NoteColumn
            NoteHead
            NoteName
            NoteSpacing
            OttavaBracket
            PaperColumn
            ParenthesesItem
            PercentRepeat
            PercentRepeatCounter
            PhrasingSlur
            PianoPedalBracket
            RehearsalMark
            RepeatSlash
            RepeatTie
            RepeatTieColumn
            Rest
            RestCollision
            Script
            ScriptColumn
            ScriptRow
            Slur
            SostenutoPedal
            SostenutoPedalLineSpanner
            SpacingSpanner
            SpanBar
            SpanBarStub
            StaffGrouper
            StaffSpacing
            StaffSymbol
            StanzaNumber
            Stem
            StemStub
            StemTremolo
            StringNumber
            StrokeFinger
            SustainPedal
            SustainPedalLineSpanner
            System
            SystemStartBar
            SystemStartBrace
            SystemStartBracket
            SystemStartSquare
            TabNoteHead
            TextScript
            TextSpanner
            Tie
            TieColumn
            TimeSignature
            TrillPitchAccidental
            TrillPitchGroup
            TrillPitchHead
            TrillSpanner
            TupletBracket
            TupletNumber
            UnaCordaPedal
            UnaCordaPedalLineSpanner
            VaticanaLigature
            VerticalAlignment
            VerticalAxisGroup
            VoiceFollower
            VoltaBracket
            VoltaBracketSpanner

        Returns tuple.
        '''
        from abjad.ly import grob_interfaces
        return sorted(grob_interfaces.keys())

    @staticmethod
    def list_known_languages():
        r'''Lists all note-input languages recognized by LilyPond parser.

        ..  container:: example

            >>> class_ = abjad.parser.LilyPondParser
            >>> for language in class_.list_known_languages():
            ...     print(language)
            ...
            catalan
            deutsch
            english
            espanol
            español
            français
            italiano
            nederlands
            norsk
            portugues
            suomi
            svenska
            vlaams

        Returns list.
        '''
        from abjad.ly import language_pitch_names
        return sorted(language_pitch_names.keys())

    @staticmethod
    def list_known_markup_functions():
        r'''Lists all markup functions recognized by LilyPond parser.

        ..  container:: example

            >>> class_  = abjad.parser.LilyPondParser
            >>> for name in class_.list_known_markup_functions():
            ...     print(name)
            ...
            abs-fontsize
            arrow-head
            auto-footnote
            backslashed-digit
            beam
            bold
            box
            bracket
            caps
            center-align
            center-column
            char
            circle
            column
            column-lines
            combine
            compound-meter
            concat
            customTabClef
            dir-column
            doubleflat
            doublesharp
            draw-circle
            draw-dashed-line
            draw-dotted-line
            draw-hline
            draw-line
            dynamic
            ellipse
            epsfile
            eyeglasses
            fermata
            fill-line
            fill-with-pattern
            filled-box
            finger
            first-visible
            flat
            fontCaps
            fontsize
            footnote
            fraction
            fret-diagram
            fret-diagram-terse
            fret-diagram-verbose
            fromproperty
            general-align
            halign
            harp-pedal
            hbracket
            hcenter-in
            hspace
            huge
            italic
            justified-lines
            justify
            justify-field
            justify-line
            justify-string
            large
            larger
            left-align
            left-brace
            left-column
            line
            lookup
            lower
            magnify
            map-markup-commands
            markalphabet
            markletter
            medium
            musicglyph
            natural
            normal-size-sub
            normal-size-super
            normal-text
            normalsize
            note
            note-by-number
            null
            number
            on-the-fly
            oval
            override
            override-lines
            pad
            pad-around
            pad-to-box
            pad-x
            page-link
            page-ref
            parenthesize
            path
            pattern
            postscript
            property-recursive
            put-adjacent
            raise
            replace
            rest
            rest-by-number
            right-align
            right-brace
            right-column
            roman
            rotate
            rounded-box
            sans
            scale
            score
            score-lines
            semiflat
            semisharp
            sesquiflat
            sesquisharp
            sharp
            simple
            slashed-digit
            small
            smallCaps
            smaller
            stencil
            strut
            sub
            super
            table-of-contents
            teeny
            text
            tied-lyric
            tiny
            translate
            translate-scaled
            transparent
            triangle
            typewriter
            underline
            upright
            vcenter
            verbatim-file
            vspace
            whiteout
            whiteout-box
            with-color
            with-dimensions
            with-link
            with-url
            woodwind-diagram
            wordwrap
            wordwrap-field
            wordwrap-internal
            wordwrap-lines
            wordwrap-string
            wordwrap-string-internal

        Returns list.
        '''
        from abjad.ly import markup_functions
        from abjad.ly import markup_list_functions
        return sorted(list(markup_functions.keys()) + list(markup_list_functions.keys()))

    @staticmethod
    def list_known_music_functions():
        r'''Lists all music functions recognized by LilyPond parser.

        ..  container:: example

            >>> class_ = abjad.parser.LilyPondParser
            >>> for name in class_.list_known_music_functions():
            ...     print(name)
            ...
            acciaccatura
            appoggiatura
            bar
            breathe
            clef
            grace
            key
            language
            makeClusters
            mark
            relative
            skip
            time
            times
            transpose

        Returns list.
        '''
        from abjad import parser as abjad_parser
        from abjad.ly import current_module
        music_functions = []
        for name in current_module:
            if not isinstance(current_module[name], dict):
                continue
            if 'type' not in current_module[name]:
                continue
            if not current_module[name]['type'] == 'ly:music-function?':
                continue
            if not hasattr(abjad_parser.GuileProxy, name):
                continue
            music_functions.append(name)
        return sorted(music_functions)

    @classmethod
    def register_markup_function(class_, name, signature):
        r'''Registers a custom markup function globally with LilyPondParser.

        >>> name = 'my-custom-markup-function'
        >>> signature = ['markup?']
        >>> class_ = abjad.parser.LilyPondParser
        >>> class_.register_markup_function(name, signature)

        >>> parser = abjad.parser.LilyPondParser()
        >>> string = r"\markup { \my-custom-markup-function { foo bar baz } }"
        >>> parser(string)
        Markup(contents=[MarkupCommand('my-custom-markup-function', ['foo', 'bar', 'baz'])])

        `signature` should be a sequence of zero or more type-predicate names,
        as understood by LilyPond.  Consult LilyPond's documentation for a
        complete list of all understood type-predicates.

        Returns none.
        '''
        from abjad.ly.markup_functions import markup_functions
        assert isinstance(name, str)
        assert all(not x.isspace() for x in name)
        assert isinstance(signature, collections.Iterable)
        for predicate in signature:
            assert isinstance(predicate, str)
            assert all(not x.isspace() for x in predicate)
            assert predicate.endswith('?')
        markup_functions[name] = tuple(signature)

    ### PUBLIC PROPERTIES ###

    @property
    def available_languages(self):
        r'''Tuple of pitch-name languages supported by LilyPondParser.

        ..  container:: example

            >>> parser = abjad.parser.LilyPondParser()
            >>> for language in parser.available_languages:
            ...     print(language)
            ...
            catalan
            deutsch
            english
            espanol
            español
            français
            italiano
            nederlands
            norsk
            portugues
            suomi
            svenska
            vlaams

        Returns tuple.
        '''
        return tuple(sorted(self._language_pitch_names.keys()))

    @property
    def default_language(self):
        r'''Gets and sets default language of parser.

        ..  container:: example

            >>> parser = abjad.parser.LilyPondParser()

            >>> parser.default_language
            'english'

            >>> parser('{ c df e fs }')
            Container('c4 df4 e4 fs4')

            >>> parser.default_language = 'nederlands'
            >>> parser.default_language
            'nederlands'

            >>> parser('{ c des e fis }')
            Container('c4 df4 e4 fs4')

        Returns string.
        '''
        return self._default_language

    @default_language.setter
    def default_language(self, argument):
        assert argument in self.available_languages
        self._default_language = argument

    @property
    def lexer_rules_object(self):
        r'''Lexer rules object of LilyPond parser.
        '''
        return self._lexdef

    @property
    def parser_rules_object(self):
        r'''Parser rules object of LilyPond parser.
        '''
        return self._syndef



lilypond_version = "2.19.24"

interface_properties = {
    "accidental-interface": [
        "alteration",
        "avoid-slur",
        "glyph-name",
        "glyph-name-alist",
        "hide-tied-accidental-after-break",
        "parenthesized",
        "restore-first",
        ],
    "accidental-placement-interface": [
        "direction",
        "padding",
        "positioning-done",
        "right-padding",
        "script-priority",
        ],
    "accidental-suggestion-interface": [],
    "align-interface": [
        "align-dir",
        "axes",
        "padding",
        "positioning-done",
        "stacking-dir",
        ],
    "ambitus-interface": [
        "gap",
        "length-fraction",
        "maximum-gap",
        "thickness",
        ],
    "arpeggio-interface": [
        "arpeggio-direction",
        "dash-definition",
        "positions",
        "protrusion",
        "script-priority",
        ],
    "axis-group-interface": [
        "axes",
        "default-staff-staff-spacing",
        "max-stretch",
        "no-alignment",
        "nonstaff-nonstaff-spacing",
        "nonstaff-relatedstaff-spacing",
        "nonstaff-unrelatedstaff-spacing",
        "staff-affinity",
        "staff-staff-spacing",
        ],
    "balloon-interface": [
        "annotation-balloon",
        "annotation-line",
        "padding",
        "text",
        ],
    "bar-line-interface": [
        "allow-span-bar",
        "bar-extent",
        "gap",
        "glyph",
        "glyph-name",
        "hair-thickness",
        "kern",
        "rounded",
        "segno-kern",
        "thick-thickness",
        ],
    "bass-figure-alignment-interface": [],
    "bass-figure-interface": [
        "implicit",
        ],
    "beam-interface": [
        "X-positions",
        "annotation",
        "auto-knee-gap",
        "beam-thickness",
        "beamed-stem-shorten",
        "beaming",
        "break-overshoot",
        "clip-edges",
        "collision-interfaces",
        "collision-voice-only",
        "concaveness",
        "damping",
        "details",
        "direction",
        "gap",
        "gap-count",
        "grow-direction",
        "inspect-quants",
        "knee",
        "length-fraction",
        "neutral-direction",
        "positions",
        "skip-quanting",
        ],
    "bend-after-interface": [
        "thickness",
        ],
    "break-alignable-interface": [
        "break-align-symbols",
        "non-break-align-symbols",
        ],
    "break-aligned-interface": [
        "break-align-anchor",
        "break-align-anchor-alignment",
        "break-align-symbol",
        "space-alist",
        ],
    "break-alignment-interface": [
        "break-align-orders",
        "positioning-done",
        ],
    "breathing-sign-interface": [
        "direction",
        ],
    "chord-name-interface": [],
    "clef-interface": [
        "full-size-change",
        "glyph",
        "glyph-name",
        "non-default",
        ],
    "clef-modifier-interface": [
        "clef-alignments",
        ],
    "cluster-beacon-interface": [
        "positions",
        ],
    "cluster-interface": [
        "padding",
        "style",
        ],
    "custos-interface": [
        "neutral-direction",
        "neutral-position",
        "style",
        ],
    "dot-column-interface": [
        "chord-dots-limit",
        "direction",
        "positioning-done",
        ],
    "dots-interface": [
        "direction",
        "dot-count",
        "style",
        ],
    "dynamic-interface": [],
    "dynamic-line-spanner-interface": [
        "avoid-slur",
        ],
    "dynamic-text-interface": [
        "right-padding",
        ],
    "dynamic-text-spanner-interface": [
        "text",
        ],
    "enclosing-bracket-interface": [
        "bracket-flare",
        "edge-height",
        "padding",
        "shorten-pair",
        "thickness",
        ],
    "episema-interface": [],
    "figured-bass-continuation-interface": [
        "padding",
        "thickness",
        ],
    "finger-interface": [],
    "fingering-column-interface": [
        "padding",
        "positioning-done",
        "snap-radius",
        ],
    "flag-interface": [
        "glyph-name",
        "stroke-style",
        "style",
        ],
    "font-interface": [
        "font-encoding",
        "font-family",
        "font-name",
        "font-series",
        "font-shape",
        "font-size",
        ],
    "footnote-interface": [
        "automatically-numbered",
        "footnote",
        "footnote-text",
        ],
    "footnote-spanner-interface": [
        "footnote-text",
        ],
    "fret-diagram-interface": [
        "align-dir",
        "dot-placement-list",
        "fret-diagram-details",
        "size",
        "thickness",
        ],
    "glissando-interface": [],
    "grace-spacing-interface": [
        "common-shortest-duration",
        ],
    "gregorian-ligature-interface": [],
    "grid-line-interface": [
        "thickness",
        ],
    "grid-point-interface": [],
    "grob-interface": [
        "X-extent",
        "X-offset",
        "Y-extent",
        "Y-offset",
        "after-line-breaking",
        "avoid-slur",
        "before-line-breaking",
        "color",
        "cross-staff",
        "extra-offset",
        "footnote-music",
        "forced-spacing",
        "horizontal-skylines",
        "id",
        "layer",
        "minimum-X-extent",
        "minimum-Y-extent",
        "parenthesis-friends",
        "rotation",
        "skyline-horizontal-padding",
        "springs-and-rods",
        "stencil",
        "transparent",
        "vertical-skylines",
        "whiteout",
        "whiteout-box",
        ],
    "hairpin-interface": [
        "bound-padding",
        "broken-bound-padding",
        "circled-tip",
        "grow-direction",
        "height",
        ],
    "hara-kiri-group-spanner-interface": [
        "remove-empty",
        "remove-first",
        "remove-layer",
        ],
    "horizontal-bracket-interface": [
        "bracket-flare",
        "connect-to-neighbor",
        "edge-height",
        "shorten-pair",
        ],
    "inline-accidental-interface": [],
    "instrument-specific-markup-interface": [
        "fret-diagram-details",
        "graphical",
        "harp-pedal-details",
        "size",
        "thickness",
        ],
    "item-interface": [
        "break-visibility",
        "extra-spacing-height",
        "extra-spacing-width",
        "non-musical",
        ],
    "key-cancellation-interface": [],
    "key-signature-interface": [
        "alteration-alist",
        "flat-positions",
        "glyph-name-alist",
        "padding",
        "padding-pairs",
        "sharp-positions",
        ],
    "kievan-ligature-interface": [
        "padding",
        ],
    "ledger-line-spanner-interface": [
        "gap",
        "length-fraction",
        "minimum-length-fraction",
        ],
    "ledgered-interface": [
        "no-ledgers",
        ],
    "ligature-bracket-interface": [
        "height",
        "thickness",
        "width",
        ],
    "ligature-head-interface": [],
    "ligature-interface": [],
    "line-interface": [
        "arrow-length",
        "arrow-width",
        "dash-fraction",
        "dash-period",
        "style",
        "thickness",
        "zigzag-length",
        "zigzag-width",
        ],
    "line-spanner-interface": [
        "bound-details",
        "extra-dy",
        "gap",
        "left-bound-info",
        "right-bound-info",
        "simple-Y",
        "thickness",
        "to-barline",
        ],
    "lyric-extender-interface": [
        "left-padding",
        "next",
        "right-padding",
        "thickness",
        ],
    "lyric-hyphen-interface": [
        "dash-period",
        "height",
        "length",
        "minimum-distance",
        "minimum-length",
        "padding",
        "thickness",
        ],
    "lyric-interface": [],
    "lyric-syllable-interface": [],
    "mark-interface": [],
    "measure-counter-interface": [
        "count-from",
        ],
    "measure-grouping-interface": [
        "height",
        "style",
        "thickness",
        ],
    "melody-spanner-interface": [
        "neutral-direction",
        ],
    "mensural-ligature-interface": [
        "thickness",
        ],
    "metronome-mark-interface": [],
    "multi-measure-interface": [
        "bound-padding",
        ],
    "multi-measure-rest-interface": [
        "bound-padding",
        "expand-limit",
        "hair-thickness",
        "measure-count",
        "minimum-length",
        "round-up-exceptions",
        "round-up-to-longer-rest",
        "spacing-pair",
        "thick-thickness",
        "usable-duration-logs",
        ],
    "note-collision-interface": [
        "merge-differently-dotted",
        "merge-differently-headed",
        "positioning-done",
        "prefer-dotted-right",
        ],
    "note-column-interface": [
        "force-hshift",
        "horizontal-shift",
        "ignore-collision",
        ],
    "note-head-interface": [
        "duration-log",
        "glyph-name",
        "ignore-ambitus",
        "note-names",
        "stem-attachment",
        "style",
        ],
    "note-name-interface": [],
    "note-spacing-interface": [
        "knee-spacing-correction",
        "same-direction-correction",
        "space-to-barline",
        "stem-spacing-correction",
        ],
    "number-interface": [
        "number-type",
        ],
    "only-prebreak-interface": [],
    "ottava-bracket-interface": [
        "bracket-flare",
        "edge-height",
        "minimum-length",
        "shorten-pair",
        ],
    "outside-staff-axis-group-interface": [
        "outside-staff-placement-directive",
        ],
    "outside-staff-interface": [
        "outside-staff-horizontal-padding",
        "outside-staff-padding",
        "outside-staff-priority",
        ],
    "paper-column-interface": [
        "between-cols",
        "full-measure-extra-space",
        "labels",
        "line-break-penalty",
        "line-break-permission",
        "line-break-system-details",
        "page-break-penalty",
        "page-break-permission",
        "page-turn-penalty",
        "page-turn-permission",
        "rhythmic-location",
        "shortest-playing-duration",
        "shortest-starter-duration",
        "used",
        "when",
        ],
    "parentheses-interface": [
        "padding",
        "stencils",
        ],
    "percent-repeat-interface": [
        "dot-negative-kern",
        "slash-negative-kern",
        "slope",
        "thickness",
        ],
    "percent-repeat-item-interface": [
        "dot-negative-kern",
        "slash-negative-kern",
        "slope",
        "thickness",
        ],
    "piano-pedal-bracket-interface": [
        "bound-padding",
        "bracket-flare",
        "edge-height",
        "shorten-pair",
        ],
    "piano-pedal-interface": [],
    "piano-pedal-script-interface": [],
    "pitched-trill-interface": [],
    "pure-from-neighbor-interface": [],
    "rest-collision-interface": [
        "minimum-distance",
        "positioning-done",
        ],
    "rest-interface": [
        "direction",
        "minimum-distance",
        "style",
        "voiced-position",
        ],
    "rhythmic-grob-interface": [],
    "rhythmic-head-interface": [
        "duration-log",
        "glissando-skip",
        ],
    "script-column-interface": [],
    "script-interface": [
        "avoid-slur",
        "positioning-done",
        "script-priority",
        "side-relative-direction",
        "slur-padding",
        "toward-stem-shift",
        "toward-stem-shift-in-column",
        ],
    "self-alignment-interface": [
        "X-align-on-main-noteheads",
        "parent-alignment-X",
        "parent-alignment-Y",
        "self-alignment-X",
        "self-alignment-Y",
        ],
    "semi-tie-column-interface": [
        "head-direction",
        "positioning-done",
        "tie-configuration",
        ],
    "semi-tie-interface": [
        "control-points",
        "details",
        "direction",
        "head-direction",
        "thickness",
        ],
    "separation-item-interface": [
        "X-extent",
        "horizontal-skylines",
        "padding",
        "skyline-vertical-padding",
        ],
    "side-position-interface": [
        "add-stem-support",
        "direction",
        "horizon-padding",
        "minimum-space",
        "padding",
        "side-axis",
        "slur-padding",
        "staff-padding",
        "use-skylines",
        ],
    "slur-interface": [
        "annotation",
        "avoid-slur",
        "control-points",
        "dash-definition",
        "details",
        "direction",
        "eccentricity",
        "height-limit",
        "inspect-index",
        "inspect-quants",
        "line-thickness",
        "positions",
        "ratio",
        "thickness",
        ],
    "spaceable-grob-interface": [
        "allow-loose-spacing",
        "keep-inside-line",
        "measure-length",
        ],
    "spacing-interface": [],
    "spacing-options-interface": [
        "shortest-duration-space",
        "spacing-increment",
        ],
    "spacing-spanner-interface": [
        "average-spacing-wishes",
        "base-shortest-duration",
        "common-shortest-duration",
        "packed-spacing",
        "shortest-duration-space",
        "spacing-increment",
        "strict-grace-spacing",
        "strict-note-spacing",
        "uniform-stretching",
        ],
    "span-bar-interface": [
        "glyph-name",
        ],
    "spanner-interface": [
        "minimum-length",
        "minimum-length-after-break",
        "normalized-endpoints",
        "spanner-id",
        "to-barline",
        ],
    "staff-grouper-interface": [
        "staff-staff-spacing",
        "staffgroup-staff-spacing",
        ],
    "staff-spacing-interface": [
        "stem-spacing-correction",
        ],
    "staff-symbol-interface": [
        "break-align-symbols",
        "ledger-extra",
        "ledger-line-thickness",
        "ledger-positions",
        "line-count",
        "line-positions",
        "staff-space",
        "thickness",
        "width",
        ],
    "staff-symbol-referencer-interface": [
        "staff-position",
        ],
    "stanza-number-interface": [],
    "stem-interface": [
        "avoid-note-head",
        "beaming",
        "beamlet-default-length",
        "beamlet-max-length-proportion",
        "default-direction",
        "details",
        "direction",
        "double-stem-separation",
        "duration-log",
        "french-beaming",
        "length",
        "length-fraction",
        "max-beam-connect",
        "neutral-direction",
        "no-stem-extend",
        "positioning-done",
        "stem-begin-position",
        "stemlet-length",
        "thickness",
        ],
    "stem-tremolo-interface": [
        "beam-thickness",
        "beam-width",
        "direction",
        "flag-count",
        "length-fraction",
        "shape",
        "slope",
        ],
    "string-number-interface": [],
    "stroke-finger-interface": [
        "digit-names",
        ],
    "system-interface": [
        "labels",
        ],
    "system-start-delimiter-interface": [
        "collapse-height",
        "style",
        "thickness",
        ],
    "system-start-text-interface": [
        "long-text",
        "self-alignment-X",
        "self-alignment-Y",
        "text",
        ],
    "tab-note-head-interface": [
        "details",
        ],
    "text-interface": [
        "baseline-skip",
        "flag-style",
        "replacement-alist",
        "text",
        "text-direction",
        "word-space",
        ],
    "text-script-interface": [
        "avoid-slur",
        "script-priority",
        ],
    "tie-column-interface": [
        "positioning-done",
        "tie-configuration",
        ],
    "tie-interface": [
        "annotation",
        "avoid-slur",
        "control-points",
        "dash-definition",
        "details",
        "direction",
        "head-direction",
        "line-thickness",
        "neutral-direction",
        "staff-position",
        "thickness",
        ],
    "time-signature-interface": [
        "fraction",
        "style",
        ],
    "trill-pitch-accidental-interface": [],
    "trill-spanner-interface": [],
    "tuplet-bracket-interface": [
        "X-positions",
        "avoid-scripts",
        "bracket-flare",
        "bracket-visibility",
        "break-overshoot",
        "connect-to-neighbor",
        "direction",
        "edge-height",
        "edge-text",
        "full-length-padding",
        "full-length-to-extent",
        "gap",
        "padding",
        "positions",
        "shorten-pair",
        "staff-padding",
        "thickness",
        ],
    "tuplet-number-interface": [
        "avoid-slur",
        "direction",
        "knee-to-beam",
        ],
    "unbreakable-spanner-interface": [
        "breakable",
        ],
    "vaticana-ligature-interface": [
        "glyph-name",
        "thickness",
        ],
    "volta-bracket-interface": [
        "height",
        "shorten-pair",
        "thickness",
        ],
    "volta-interface": [],
    }
